# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from functools import reduce

import numpy as np

from .domains.domain import Domain


class DomainTuple:
    """Ordered sequence of Domain objects.

    This class holds a tuple of :class:`Domain` objects, which together form
    the space on which a :class:`Field` is defined.
    This corresponds to a tensor product of the corresponding vector
    fields.

    Notes
    -----

    DomainTuples should never be created using the constructor, but rather
    via the factory function :attr:`make`!
    """
    _tupleCache = {}
    _scalarDomain = None

    def __init__(self, domain, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        self._dom = self._parse_domain(domain)
        self._axtuple = self._get_axes_tuple()
        self._shape = reduce(lambda x, y: x+y, (sp.shape for sp in self._dom),
                             ())
        self._size = reduce(lambda x, y: x*y, self._shape, 1)

    def _get_axes_tuple(self):
        i = 0
        res = [None]*len(self._dom)
        for idx, thing in enumerate(self._dom):
            nax = len(thing.shape)
            res[idx] = tuple(range(i, i+nax))
            i += nax
        return tuple(res)

    @staticmethod
    def make(domain):
        """Returns a DomainTuple matching `domain`.

        This function checks whether a matching DomainTuple already exists.
        If yes, this object is returned, otherwise a new DomainTuple object
        is created and returned.

        Parameters
        ----------
        domain : Domain or tuple of Domain or DomainTuple
            The geometrical structure for which the DomainTuple shall be
            obtained.
        """
        if isinstance(domain, DomainTuple):
            return domain
        from .multi_domain import MultiDomain
        if isinstance(domain, MultiDomain):
            raise TypeError("Cannot create DomainTuple from MultiDomain")
        domain = DomainTuple._parse_domain(domain)
        obj = DomainTuple._tupleCache.get(domain)
        if obj is not None:
            return obj
        obj = DomainTuple(domain, _callingfrommake=True)
        DomainTuple._tupleCache[domain] = obj
        return obj

    @staticmethod
    def _parse_domain(domain):
        if domain is None:
            return ()
        if isinstance(domain, Domain):
            return (domain,)

        if not isinstance(domain, tuple):
            domain = tuple(domain)
        for d in domain:
            if not isinstance(d, Domain):
                raise TypeError(
                    "Given object contains something that is not an "
                    "instance of Domain class.")
        return domain

    def __getitem__(self, i):
        return self._dom[i]

    @property
    def shape(self):
        """tuple of int: number of pixels along each axis

        The shape of the array-like object required to store information
        defined on the DomainTuple.
        """
        return self._shape

    @property
    def size(self):
        """int : total number of pixels.

        Equivalent to the products over all entries in the object's shape.
        """
        return self._size

    def scalar_weight(self, spaces=None):
        """Returns the uniform volume element for a sub-domain of `self`.

        Parameters
        ----------
        spaces : int, tuple of int or None
            Indices of the sub-domains to be considered.
            If `None`, the entire domain is used.

        Returns
        -------
        float or None
            If the requested sub-domain has a uniform volume element, it is
            returned. Otherwise, `None` is returned.
        """
        if np.isscalar(spaces):
            return self._dom[spaces].scalar_dvol

        if spaces is None:
            spaces = range(len(self._dom))
        res = 1.
        for i in spaces:
            tmp = self._dom[i].scalar_dvol
            if tmp is None:
                return None
            res *= tmp
        return res

    def total_volume(self, spaces=None):
        """Returns the total volume of `self` or of a subspace of it.

        Parameters
        ----------
        spaces : int, tuple of int or None
            Indices of the sub-domains of the domain to be considered.
            If `None`, the total volume of the whole domain is returned.

        Returns
        -------
        float
            the total volume of the requested (sub-)domain.
        """
        if np.isscalar(spaces):
            return self._dom[spaces].total_volume

        if spaces is None:
            spaces = range(len(self._dom))
        res = 1.
        for i in spaces:
            res *= self._dom[i].total_volume
        return res

    @property
    def axes(self):
        """tuple of tuple of int : axis indices of the underlying domains"""
        return self._axtuple

    def __len__(self):
        return len(self._dom)

    def __hash__(self):
        return hash(self._dom)

    def __eq__(self, x):
        return (self is x) or (isinstance(x, DomainTuple) and self._dom == x._dom)

    def __ne__(self, x):
        return not self.__eq__(x)

    def __str__(self):
        if len(self) == 0:
            return "DomainTuple, len: 0"
        return ("DomainTuple, len: {}\n".format(len(self)) +
                "\n".join("* " + str(i) for i in self))

    def __reduce__(self):
        return (_unpickleDomainTuple, (self._dom,))

    @staticmethod
    def scalar_domain():
        if DomainTuple._scalarDomain is None:
            DomainTuple._scalarDomain = DomainTuple.make(())
        return DomainTuple._scalarDomain

    def __repr__(self):
        return f"DomainTuple.make({tuple(self)})"


def _unpickleDomainTuple(*args):
    return DomainTuple.make(*args)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from dataclasses import field
from typing import Callable, Union

import jax.numpy as jnp
from jax import eval_shape
from jax.tree_util import Partial
from numpy import typing as npt

from ..model import Model, WrappedCall
from ..prior import NormalPrior
from ..tree_math import ShapeWithDtype
from .grid import Grid
from .kernel import ICRKernel, Kernel, apply_kernel


class ICRField(Model):
    grid: Grid
    kernel: Kernel
    covariance: Union[Model, Callable] = field(metadata=dict(static=False))
    offset: Model = field(metadata=dict(static=False))
    compress: bool
    fixed_kernel: bool

    def __init__(
        self,
        grid: Grid,
        *,
        kernel: Union[
            dict, Model, Callable[[npt.NDArray, npt.NDArray], npt.NDArray]
        ] = dict(kind="experimental_matern"),
        offset=0.0,
        window_size=None,
        compress: Union[bool, dict] = dict(
            rtol=1e-5, atol=1e-10, buffer_size=10_000, use_distances=True
        ),
        prefix="mgcfm",
    ):
        """Correlated field model utilizing ICR to work on arbitrarily charted grids.

        Parameters
        ----------
        grid: Grid
            Storage object with instructions on how pixels are spaced.
        kernel: dict or Model or callable
            Parameters for the kernel. Currently supports `kind='experimental_matern'`
            with the following additional arguments

            - scale: tuple or Model or float -- Prior scale of the kernel.
            - cutoff: tuple or Model or float -- Prior cutoff of the covariance kernel.
            - loglogslope: tuple or Model or float -- Prior power-law slope of the covariance kernel correlating modes.
            - n_integrate = 2_000 -- Number of integration points for the harmonic transformation.
            - n_interpolate = 512 -- Number of interpolation points for interpolating the kernel.
            - interpolation_dists_min_max -- Interpolation range. Automatically chosen based on cutoff if left unspecified.
            - integration_dists_min_max -- Integration range for the harmonic transform. Automatically chosen based on cutoff if left unspecified.

            alternatively, `kernel` can be a NIFTy `Model` which yields a callable
            covariance function (`kernel(xi: dict) -> callable[[x, y], z]`), or a
            `callable[[x, y], z]` covarinace function.
        offset: tuple or callable or float
            Prior shift from zero in addition to the field intrinsic random shift.
        window_size:
            Shape of the refinement matrices. Its value is automatically inferred
            from the padding in the grid or set to sane defaults if left unspecified.
        compress: dict
            Arguments for compressing the refinement matrices. If this is set to an
            empty dictionary, no compression is applied.
        prexi: str
            Prefix for the latent parameter names.
        """
        from .matern import MaternHarmonicCovariance

        self.grid = grid
        shapes = [
            ShapeWithDtype(self.grid.at(lvl).shape, jnp.float_)
            for lvl in range(grid.depth + 1)
        ]
        self._name_exc = str(prefix) + "excitations"
        domain = {self._name_exc: shapes}

        # Parse kernel
        DEFAULT_KERNEL_KIND = "experimental_matern"
        fixed_kernel, covariance = False, None
        if isinstance(kernel, dict):
            kernel = kernel.copy()
            kernel_kind = kernel.pop("kind", DEFAULT_KERNEL_KIND).lower()
            if kernel_kind == "experimental_matern":
                coord_swd = eval_shape(
                    grid.at(0).index2coord, ShapeWithDtype((grid.at(0).ndim,), jnp.int_)
                )
                kernel.setdefault("ndim", coord_swd.shape[0])
                covariance = MaternHarmonicCovariance(**kernel, prefix=prefix)
            else:
                raise ValueError(f"kernel {kernel_kind!r} not supported")
        elif isinstance(kernel, Model):
            covariance = kernel
        elif callable(kernel):
            fixed_kernel = True
            covariance = Partial(kernel)
        else:
            raise TypeError(f"invalid kernel type; got {kernel!r}")
        self.fixed_kernel = fixed_kernel
        self.covariance = covariance
        if not self.fixed_kernel:
            domain |= self.covariance.domain

        # Parse offset
        name_off = prefix + "offset"
        if isinstance(offset, (tuple, list)):
            offset = NormalPrior(*offset, name=name_off)
        elif callable(offset) and not isinstance(offset, Model):
            offset = WrappedCall(offset, name=name_off, white_init=True)
        if not (isinstance(offset, Model) or isinstance(offset, float)):
            raise ValueError(f"invalid `offset`; got {offset!r}")
        if isinstance(offset, Model):
            domain |= offset.domain
        self.offset = offset

        self.compress = isinstance(compress, dict) and len(compress) > 0
        kernel = ICRKernel(self.grid, None, window_size=window_size)
        if self.compress:
            kernel = kernel.compress_indices(**compress)
        if self.fixed_kernel:
            kernel = kernel.replace(covariance=self.covariance)
            if self.compress:
                kernel = kernel.compress_matrices()
        self.kernel = kernel

        # TODO: join init instead of domain
        super().__init__(domain=domain, white_init=True)

    def __call__(self, x):
        if not self.fixed_kernel:
            kernel = self.kernel.replace(covariance=self.covariance(x))
            if self.compress:
                kernel = kernel.compress_matrices()
        off = self.offset(x) if isinstance(self.offset, Model) else self.offset
        return off + apply_kernel(x[self._name_exc], kernel=kernel)[-1]



# %%

import jax
import matplotlib.pyplot as plt
import nifty.re as jft
from jax import random

jax.config.update("jax_enable_x64", True)

seed = 42
key = random.PRNGKey(seed)

# %% [markdown]
# ## ICR Correlated Field Model
#
# Create a correlated field on an arbitrary grid. There are a few prespecfieid grids
# but it is trivial to define a new grid on top of e.g. `SimpleOpenGrid`. NIFTy
# provides grids for common spaces such as a HEALPIx grid, a logarithmic grid, and a
# `HPLogRGrid` (HEALPix times log radius grid).

# %%
shape = (32, 32)

grid = jft.SimpleOpenGrid(min_shape=shape)
cf = jft.ICRField(
    grid,
    kernel=dict(
        kind="experimental_matern",
        scale=(1.0, 0.2),
        cutoff=(1e-1, 1e-2),
        loglogslope=(-3.0, 0.5),
    ),
)


# %% [markdown]
# ## The likelihood

# %%
signal_response = signal = cf
noise_cov = lambda x: 0.1**2 * x
noise_cov_inv = lambda x: 0.1**-2 * x

# Create synthetic data
key, subkey = random.split(key)
pos_truth = jft.random_like(subkey, signal_response.domain)
signal_response_truth = signal_response(pos_truth)
key, subkey = random.split(key)
noise_truth = (
    (noise_cov(jft.ones_like(signal_response.target))) ** 0.5
) * jft.random_like(key, signal_response.target)
data = signal_response_truth + noise_truth

lh = jft.Gaussian(data, noise_cov_inv).amend(signal_response)

# %% [markdown]
# ## The inference

# %%
n_vi_iterations = 6
delta = 1e-4
n_samples = 2

key, k_i, k_o = random.split(key, 3)
# NOTE, changing the number of samples always triggers a resampling
samples, state = jft.optimize_kl(
    lh,
    jft.Vector(lh.init(k_i)),
    n_total_iterations=n_vi_iterations,
    n_samples=n_samples,
    # Source for the stochasticity for sampling
    key=k_o,
    # point_estimates=("mgcfmcutoff",),
    # Arguments for the conjugate gradient method used to drawing samples
    draw_linear_kwargs=dict(
        cg_name="SL",
        cg_kwargs=dict(absdelta=delta * jft.size(lh.domain) / 10.0, maxiter=100),
    ),
    # Arguments for the minimizer of the KL-divergence cost potential
    kl_kwargs=dict(
        minimize_kwargs=dict(
            name="M", xtol=delta, cg_kwargs=dict(name=None), maxiter=35
        )
    ),
    sample_mode="linear_resample",
    odir="results_icr",
    resume=False,
)

# %%
post_sr_mean = jft.mean(tuple(signal(s) for s in samples))
to_plot = [
    ("Signal", signal(pos_truth), "im"),
    ("Noise", noise_truth, "im"),
    ("Data", data, "im"),
    ("Reconstruction", post_sr_mean, "im"),
]
fig, axs = plt.subplots(2, 3, figsize=(16, 9))
for ax, v in zip(axs.flat, to_plot):
    title, field, tp, *labels = v
    ax.set_title(title)
    if tp == "im":
        end = grid.at(-1).shape  # grid can be arbitrarily deformed
        im = ax.imshow(field.T, cmap="inferno", extent=(0.0, end[0], 0.0, end[1]))
        plt.colorbar(im, ax=ax, orientation="horizontal")
    else:
        ax_plot = ax.loglog if tp == "loglog" else ax.plot
        x = field[0]
        for f in field[1:]:
            ax_plot(x, f, alpha=0.7)
for ax in axs.flat[len(to_plot) :]:
    ax.set_axis_off()
fig.tight_layout()
fig.savefig("results_icr_full_reconstruction.png", dpi=400)
plt.show()

# %%



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from datetime import datetime
from functools import partial
from typing import Any, Callable, NamedTuple, Optional, Tuple, Union

import jax
from jax import numpy as jnp

from .logger import logger
from .misc import doc_from, safeguard_arguments_against_accidental_calls_into_jax
from .tree_math import assert_arithmetics, result_type, size, vdot, where, zeros_like
from .tree_math import norm as jft_norm

HessVP = Callable[[jnp.ndarray], jnp.ndarray]

N_RESET = 20


class CGResults(NamedTuple):
    x: jnp.ndarray
    nit: Union[int, jnp.ndarray]
    nfev: Union[int, jnp.ndarray]  # number of matrix-evaluations
    info: Union[int, jnp.ndarray]
    success: Union[bool, jnp.ndarray]


def cg(mat, j, x0=None, *args, **kwargs) -> Tuple[Any, Union[int, jnp.ndarray]]:
    """Solve `mat(x) = j` using Conjugate Gradient. `mat` must be callable and
    represent a hermitian, positive definite matrix.

    Notes
    -----
    If set, the parameters `absdelta` and `resnorm` always take precedence over
    `tol` and `atol`.
    """
    assert_arithmetics(j)
    if x0 is not None:
        assert_arithmetics(x0)
    cg_res = _cg(mat, j, x0, *args, **kwargs)
    return cg_res.x, cg_res.info


@doc_from(cg)
def static_cg(mat, j, x0=None, *args, **kwargs):
    assert_arithmetics(j)
    if x0 is not None:
        assert_arithmetics(x0)
    cg_res = _static_cg(mat, j, x0, *args, **kwargs)
    return cg_res.x, cg_res.info


@safeguard_arguments_against_accidental_calls_into_jax
def _cg_pretty_print_it(
    name,
    i,
    *,
    energy,
    energy_diff,
    absdelta=None,
    norm=None,
    resnorm=None,
    maxiter=None,
):
    if maxiter is not None and i == maxiter:
        i_str = "✖" * len(str(i)) + f" ({i})"
    else:
        i_str = str(i)
    msg = f"{name}: Iteration {i_str} ⛰:{energy:+.4e} Δ⛰:{energy_diff:.4e}"
    msg += f" ➽:{absdelta:.4e}" if absdelta is not None else ""
    if norm is not None and resnorm is not None:
        msg += f" |∇|:{norm:.4e} ➽:{resnorm:.4e}"
    logger.info(msg)


# Taken from nifty
def _cg(
    mat,
    j,
    x0=None,
    *,
    absdelta=None,
    resnorm=None,
    norm_ord=None,
    tol=1e-5,  # taken from SciPy's linalg.cg
    atol=0.0,
    miniter=None,
    maxiter=None,
    name=None,
    time_threshold=None,
    _raise_nonposdef=True,
) -> CGResults:
    norm_ord = 2 if norm_ord is None else norm_ord  # TODO: change to 1
    maxiter_fallback = 20 * size(j)  # taken from SciPy's NewtonCG minimzer
    miniter = (
        min((6, maxiter if maxiter is not None else maxiter_fallback))
        if miniter is None
        else miniter
    )
    maxiter = (
        max((min((200, maxiter_fallback)), miniter)) if maxiter is None else maxiter
    )

    if absdelta is None and resnorm is None:  # fallback convergence criterion
        resnorm = jnp.maximum(tol * jft_norm(j, ord=norm_ord), atol)

    common_dtp = result_type(j)
    eps = 6.0 * jnp.finfo(common_dtp).eps  # taken from SciPy's NewtonCG minimzer
    tiny = 6.0 * jnp.finfo(common_dtp).tiny

    if x0 is None:
        pos = zeros_like(j)
        r = -j
        d = r
        # energy = .5xT M x - xT j
        energy = 0.0
        nfev = 0
    else:
        pos = x0
        r = mat(pos) - j
        d = r
        energy = float(vdot((r - j) / 2, pos))
        nfev = 1
    previous_gamma = float(vdot(r, r))

    info = -1
    i = 0
    energy_diff = jnp.inf
    norm = None
    pp = partial(
        _cg_pretty_print_it, name, absdelta=absdelta, resnorm=resnorm, maxiter=maxiter
    )
    if name is not None:
        if resnorm is not None:
            norm = jft_norm(r, ord=norm_ord)
        else:
            norm = None
        pp(i, energy=energy, energy_diff=energy_diff, norm=norm)

    if previous_gamma == 0:
        info = 0
        return CGResults(x=pos, info=info, nit=0, nfev=nfev, success=True)

    for i in range(1, maxiter + 1):
        q = mat(d)
        nfev += 1

        curv = float(vdot(d, q))
        if curv == 0.0:
            if _raise_nonposdef:
                nm = "CG" if name is None else name
                raise ValueError(f"{nm}: zero curvature")
            info = 0
            break
        elif curv < 0.0:
            if _raise_nonposdef:
                nm = "CG" if name is None else name
                raise ValueError(f"{nm}: negative curvature")
            if i > 1:
                info = 0
                break
            else:
                pos = previous_gamma / (-curv) * (-j)
                info = 0
                break
        alpha = previous_gamma / curv
        pos = pos - alpha * d
        if i % N_RESET == 0:
            r = mat(pos) - j
            nfev += 1
        else:
            r = r - q * alpha
        gamma = float(vdot(r, r))
        if time_threshold is not None and datetime.now() > time_threshold:
            info = i
            break
        if gamma >= 0.0 and gamma <= tiny:
            nm = "CG" if name is None else name
            logger.warning(f"{nm}: gamma=0, converged!")
            info = 0
            break
        if resnorm is not None:
            norm = float(jft_norm(r, ord=norm_ord))
            if norm < resnorm and i >= miniter:
                info = 0
                break
        else:
            norm = None
        new_energy = float(vdot((r - j) / 2, pos))
        energy_diff = energy - new_energy
        neg_energy_eps = -eps * jnp.abs(new_energy)
        if energy_diff < neg_energy_eps:
            nm = "CG" if name is None else name
            if _raise_nonposdef:
                raise ValueError(f"{nm}: WARNING: energy increased")
            logger.error(f"{nm}: WARNING: energy increased")
            info = i
            break
        if absdelta is not None and energy_diff < absdelta and i >= miniter:
            info = 0
            break
        energy = new_energy
        d = d * max(0, gamma / previous_gamma) + r
        previous_gamma = gamma

        if name is not None:
            pp(i, energy=energy, energy_diff=energy_diff, norm=norm)

    if name is not None and info != -1:
        # only print if loop was terminated via `break` otherwise everything is
        pp(i, energy=energy, energy_diff=energy_diff, norm=norm)

    info = i if info == -1 else info
    return CGResults(x=pos, info=info, nit=i, nfev=nfev, success=info == 0)


def _static_cg(
    mat,
    j,
    x0=None,
    *,
    absdelta=None,
    resnorm=None,
    norm_ord=None,
    tol=1e-5,  # taken from SciPy's linalg.cg
    atol=0.0,
    miniter=None,
    maxiter=None,
    name=None,
    _raise_nonposdef=True,
    **kwargs,
) -> CGResults:
    from jax.debug import callback
    from jax.lax import cond, while_loop

    # While in general it is strongly discouraged to use `jit` inside a
    # function, we do it here to avoid the overhead of jitting through `mat`
    # multiple times within the CG loop. This is safe because the call to
    # `while_loop` in CG implies a JIT anyways.
    mat = jax.jit(mat)

    norm_ord = 2 if norm_ord is None else norm_ord  # TODO: change to 1
    maxiter_fallback = 20 * size(j)  # taken from SciPy's NewtonCG minimzer
    miniter = (
        jnp.minimum(6, maxiter if maxiter is not None else maxiter_fallback)
        if miniter is None
        else miniter
    )
    maxiter = (
        jnp.maximum(jnp.minimum(200, maxiter_fallback), miniter)
        if maxiter is None
        else maxiter
    )

    if absdelta is None and resnorm is None:  # fallback convergence criterion
        resnorm = jnp.maximum(tol * jft_norm(j, ord=norm_ord), atol)

    common_dtp = result_type(j)
    eps = 6.0 * jnp.finfo(common_dtp).eps  # taken from SciPy's NewtonCG minimzer
    tiny = 6.0 * jnp.finfo(common_dtp).tiny

    def pp(arg):
        _cg_pretty_print_it(name, **arg)

    def continue_condition(v):
        return v["info"] < -1

    def cg_single_step(v):
        info = v["info"]
        pos, r, d, i = v["pos"], v["r"], v["d"], v["iteration"]
        previous_gamma, previous_energy = v["gamma"], v["energy"]

        i += 1

        q = mat(d)
        curv = vdot(d, q)
        alpha = previous_gamma / curv
        # ValueError("implausible or zero curvature in conjugate gradient")
        info = jnp.where(curv <= 0.0, jnp.where(_raise_nonposdef, -1, 0), info)
        alpha = jnp.where((curv <= 0.0) & (not _raise_nonposdef), 0.0, alpha)
        pos = pos - alpha * d
        pos = where(
            (curv < 0.0) & (not _raise_nonposdef) & (i <= 1),
            previous_energy / (-curv) * (-j),
            pos,
        )
        r = cond(
            (i % N_RESET == 0) & (info < -1),
            lambda x: mat(x["pos"]) - x["j"],
            lambda x: x["r"] - x["q"] * x["alpha"],
            {"pos": pos, "j": j, "r": r, "q": q, "alpha": alpha},
        )
        gamma = vdot(r, r)

        info = jnp.where((gamma >= 0.0) & (gamma <= tiny) & (info != -1), 0, info)
        if resnorm is not None:
            norm = jft_norm(r, ord=norm_ord)
            info = jnp.where((norm < resnorm) & (i >= miniter) & (info != -1), 0, info)
        else:
            norm = None
        energy = vdot((r - j) / 2, pos)
        energy_diff = previous_energy - energy
        neg_energy_eps = -eps * jnp.abs(energy)
        # print(f"energy increased", file=sys.stderr)
        info = jnp.where(
            energy_diff < neg_energy_eps,
            jnp.where(_raise_nonposdef, -1, i),
            info,
        )
        if absdelta is not None:
            info = jnp.where(
                (energy_diff < absdelta) & (i >= miniter) & (info != -1), 0, info
            )
        info = jnp.where((i >= maxiter) & (info != -1), i, info)

        d = d * jnp.maximum(0, gamma / previous_gamma) + r

        if name is not None:
            printable_state = {
                "i": i,
                "energy": energy,
                "energy_diff": energy_diff,
                "absdelta": absdelta,
                "norm": norm,
                "resnorm": resnorm,
                "maxiter": maxiter,
            }
            callback(pp, printable_state)

        ret = {
            "info": info,
            "pos": pos,
            "r": r,
            "d": d,
            "iteration": i,
            "gamma": gamma,
            "energy": energy,
        }
        return ret

    if x0 is None:
        pos = zeros_like(j)
        r = -j
        d = r
        nfev = 0
    else:
        pos = x0
        r = mat(pos) - j
        d = r
        nfev = 1
    # energy = .5xT M x - xT j
    energy = jnp.array(0.0) if x0 is None else vdot((r - j) / 2, pos)

    gamma = vdot(r, r)
    val = {
        "info": jnp.array(-2, dtype=int),
        "pos": pos,
        "r": r,
        "d": d,
        "iteration": jnp.array(0),
        "gamma": gamma,
        "energy": energy,
    }
    # Finish early if already converged in the initial iteration
    val["info"] = jnp.where(gamma == 0.0, 0, val["info"])

    if name is not None:
        if resnorm is not None:
            norm = jft_norm(r, ord=norm_ord)
        else:
            norm = None
        printable_state = {
            "i": 0,
            "energy": energy,
            "energy_diff": jnp.inf,
            "absdelta": absdelta,
            "norm": norm,
            "resnorm": resnorm,
            "maxiter": maxiter,
        }
        callback(pp, printable_state)

    val = while_loop(continue_condition, cg_single_step, val)

    i = val["iteration"]
    info = val["info"]
    nfev += i + i // N_RESET
    return CGResults(x=val["pos"], info=info, nit=i, nfev=nfev, success=info == 0)


# The following is code adapted from Nicholas Mancuso to work with pytrees
class _QuadSubproblemResult(NamedTuple):
    step: jnp.ndarray
    hits_boundary: Union[bool, jnp.ndarray]
    pred_f: Union[float, jnp.ndarray]
    nit: Union[int, jnp.ndarray]
    nfev: Union[int, jnp.ndarray]
    njev: Union[int, jnp.ndarray]
    nhev: Union[int, jnp.ndarray]
    success: Union[bool, jnp.ndarray]


class _CGSteihaugState(NamedTuple):
    z: jnp.ndarray
    r: jnp.ndarray
    d: jnp.ndarray
    step: jnp.ndarray
    energy: Union[None, float, jnp.ndarray]
    hits_boundary: Union[bool, jnp.ndarray]
    done: Union[bool, jnp.ndarray]
    nit: Union[int, jnp.ndarray]
    nhev: Union[int, jnp.ndarray]


def second_order_approx(
    p: jnp.ndarray,
    cur_val: Union[float, jnp.ndarray],
    g: jnp.ndarray,
    hessp_at_xk: HessVP,
) -> Union[float, jnp.ndarray]:
    return cur_val + vdot(g, p) + 0.5 * vdot(p, hessp_at_xk(p))


def get_boundaries_intersections(
    z: jnp.ndarray, d: jnp.ndarray, trust_radius: Union[float, jnp.ndarray]
):  # Adapted from SciPy
    """Solve the scalar quadratic equation ||z + t d|| == trust_radius.

    This is like a line-sphere intersection.

    Return the two values of t, sorted from low to high.
    """
    a = vdot(d, d)
    b = 2 * vdot(z, d)
    c = vdot(z, z) - trust_radius**2
    sqrt_discriminant = jnp.sqrt(b * b - 4 * a * c)

    # The following calculation is mathematically
    # equivalent to:
    # ta = (-b - sqrt_discriminant) / (2*a)
    # tb = (-b + sqrt_discriminant) / (2*a)
    # but produce smaller round off errors.
    # Look at Matrix Computation p.97
    # for a better justification.
    aux = b + jnp.copysign(sqrt_discriminant, b)
    ta = -aux / (2 * a)
    tb = -2 * c / aux

    ra, rb = where(ta < tb, (ta, tb), (tb, ta))
    return (ra, rb)


def _cg_steihaug_subproblem(
    cur_val: Union[float, jnp.ndarray],
    g: jnp.ndarray,
    hessp_at_xk: HessVP,
    *,
    trust_radius: Union[float, jnp.ndarray],
    tr_norm_ord: Union[None, int, float, jnp.ndarray] = None,
    resnorm: Optional[float],
    absdelta: Optional[float] = None,
    norm_ord: Union[None, int, float, jnp.ndarray] = None,
    miniter: Union[None, int] = None,
    maxiter: Union[None, int] = None,
    name=None,
) -> _QuadSubproblemResult:
    """
    Solve the subproblem using a conjugate gradient method.

    Parameters
    ----------
    cur_val : Union[float, jnp.ndarray]
      Objective value evaluated at the current state.
    g : jnp.ndarray
      Gradient value evaluated at the current state.
    hessp_at_xk: Callable
      Function that accepts a proposal vector and computes the result of a
      Hessian-vector product.
    trust_radius : float
      Upper bound on how large a step proposal can be.
    tr_norm_ord : {non-zero int, inf, -inf, ‘fro’, ‘nuc’}, optional
      Order of the norm for computing the length of the next step.
    norm_ord : {non-zero int, inf, -inf, ‘fro’, ‘nuc’}, optional
      Order of the norm for testing convergence.

    Returns
    -------
    result : _QuadSubproblemResult
      Contains the step proposal, whether it is at radius boundary, and
      meta-data regarding function calls and successful convergence.

    Notes
    -----
    This is algorithm (7.2) of Nocedal and Wright 2nd edition.
    Only the function that computes the Hessian-vector product is required.
    The Hessian itself is not required, and the Hessian does
    not need to be positive semidefinite.
    """
    from jax.debug import callback
    from jax.lax import switch, while_loop

    tr_norm_ord = jnp.inf if tr_norm_ord is None else tr_norm_ord  # taken from JAX
    norm_ord = 2 if norm_ord is None else norm_ord  # TODO: change to 1
    maxiter_fallback = 20 * size(g)  # taken from SciPy's NewtonCG minimzer
    miniter = (
        jnp.minimum(6, maxiter if maxiter is not None else maxiter_fallback)
        if miniter is None
        else miniter
    )
    maxiter = (
        jnp.maximum(jnp.minimum(200, maxiter_fallback), miniter)
        if maxiter is None
        else maxiter
    )

    common_dtp = result_type(g)
    eps = 6.0 * jnp.finfo(common_dtp).eps  # Inspired by SciPy's NewtonCG minimzer

    # second-order Taylor series approximation at the current values, gradient,
    # and hessian
    soa = partial(second_order_approx, cur_val=cur_val, g=g, hessp_at_xk=hessp_at_xk)

    @safeguard_arguments_against_accidental_calls_into_jax
    def pp(arg):
        msg = (
            "{name}: |∇|:{r_norm:.6e} ➽:{resnorm:.6e} ↗:{tr:.6e}"
            " ☞:{case:1d} #∇²:{nhev:02d}"
            "\n{name}: Iteration {i} ⛰:{energy:+.6e} Δ⛰:{energy_diff:.6e}"
            + (" ➽:{absdelta:.6e}" if arg["absdelta"] is not None else "")
            + (
                "\n{name}: Iteration Limit Reached"
                if arg["i"] == arg["maxiter"]
                else ""
            )
        )
        logger.info(msg.format(name=name, **arg))

    # helpers for internal switches in the main CGSteihaug logic
    def noop(
        param: Tuple[_CGSteihaugState, Union[float, jnp.ndarray]],
    ) -> _CGSteihaugState:
        iterp, z_next = param
        return iterp

    def step1(
        param: Tuple[_CGSteihaugState, Union[float, jnp.ndarray]],
    ) -> _CGSteihaugState:
        iterp, z_next = param
        z, d, nhev = iterp.z, iterp.d, iterp.nhev

        ta, tb = get_boundaries_intersections(z, d, trust_radius)
        pa = z + ta * d
        pb = z + tb * d
        p_boundary = where(soa(pa) < soa(pb), pa, pb)
        return iterp._replace(
            step=p_boundary, nhev=nhev + 2, hits_boundary=True, done=True
        )

    def step2(
        param: Tuple[_CGSteihaugState, Union[float, jnp.ndarray]],
    ) -> _CGSteihaugState:
        iterp, z_next = param
        z, d = iterp.z, iterp.d

        ta, tb = get_boundaries_intersections(z, d, trust_radius)
        p_boundary = z + tb * d
        return iterp._replace(step=p_boundary, hits_boundary=True, done=True)

    def step3(
        param: Tuple[_CGSteihaugState, Union[float, jnp.ndarray]],
    ) -> _CGSteihaugState:
        iterp, z_next = param
        return iterp._replace(step=z_next, hits_boundary=False, done=True)

    # initialize the step
    p_origin = zeros_like(g)

    # init the state for the first iteration
    z = p_origin
    r = g
    d = -r
    energy = 0.0
    init_param = _CGSteihaugState(
        z=z,
        r=r,
        d=d,
        step=p_origin,
        energy=energy,
        hits_boundary=False,
        done=maxiter == 0,
        nit=0,
        nhev=0,
    )

    # Search for the min of the approximation of the objective function.
    def body_f(iterp: _CGSteihaugState) -> _CGSteihaugState:
        z, r, d = iterp.z, iterp.r, iterp.d
        energy, nit = iterp.energy, iterp.nit

        nit += 1

        Bd = hessp_at_xk(d)
        dBd = vdot(d, Bd)

        r_squared = vdot(r, r)
        alpha = r_squared / dBd
        z_next = z + alpha * d

        r_next = r + alpha * Bd
        r_next_squared = vdot(r_next, r_next)

        beta_next = r_next_squared / r_squared
        d_next = -r_next + beta_next * d

        accept_z_next = nit >= maxiter
        if norm_ord == 2:
            r_next_norm = jnp.sqrt(r_next_squared)
        else:
            r_next_norm = jft_norm(r_next, ord=norm_ord)
        accept_z_next |= r_next_norm < resnorm
        # Relative to a plain CG, `z_next` is negative
        energy_next = vdot((r_next + g) / 2, z_next)
        energy_diff = energy - energy_next
        if absdelta is not None:
            neg_energy_eps = -eps * jnp.abs(energy)
            accept_z_next |= (
                (energy_diff >= neg_energy_eps)
                & (energy_diff < absdelta)
                & (nit >= miniter)
            )

        # include a junk switch to catch the case where none should be executed
        z_next_norm = jft_norm(z_next, ord=tr_norm_ord)
        index = jnp.argmax(
            jnp.array([False, dBd <= 0, z_next_norm >= trust_radius, accept_z_next])
        )
        iterp = switch(index, [noop, step1, step2, step3], (iterp, z_next))

        iterp = iterp._replace(
            z=z_next,
            r=r_next,
            d=d_next,
            energy=energy_next,
            nhev=iterp.nhev + 1,
            nit=nit,
        )
        if name is not None:
            printable_state = {
                "i": nit,
                "energy": iterp.energy,
                "energy_diff": energy_diff,
                "absdelta": absdelta,
                "tr": trust_radius,
                "r_norm": r_next_norm,
                "resnorm": resnorm,
                "nhev": iterp.nhev,
                "case": index,
                "maxiter": maxiter,
            }
            callback(pp, printable_state)

        return iterp

    def cond_f(iterp: _CGSteihaugState) -> bool:
        return jnp.logical_not(iterp.done)

    # perform inner optimization to solve the constrained
    # quadratic subproblem using cg
    result = while_loop(cond_f, body_f, init_param)

    pred_f = soa(result.step)
    result = _QuadSubproblemResult(
        step=result.step,
        hits_boundary=result.hits_boundary,
        pred_f=pred_f,
        nit=result.nit,
        nfev=0,
        njev=0,
        nhev=result.nhev + 1,
        success=True,
    )

    return result



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

###############################################################################
# Compute a Wiener filter solution with NIFTy
# Shows how measurement gaps are filled in
# 1D (set mode=0), 2D (mode=1), or on the sphere (mode=2)
###############################################################################

import sys

import numpy as np

import nifty.cl as ift


def make_checkerboard_mask(position_space):
    # Checkerboard mask for 2D mode
    mask = np.ones(position_space.shape)
    for i in range(4):
        for j in range(4):
            if (i + j) % 2 == 0:
                mask[i*128//4:(i + 1)*128//4, j*128//4:(j + 1)*128//4] = 0
    return mask


def make_random_mask(domain):
    # Random mask for spherical mode
    mask = ift.from_random(domain, 'pm1')
    mask = (mask + 1)/2
    return mask.asnumpy()


def main():
    # Choose space on which the signal field is defined
    if len(sys.argv) == 2:
        mode = int(sys.argv[1])
    else:
        mode = 1

    if mode == 0:
        # One-dimensional regular grid
        position_space = ift.RGSpace([1024])
        mask = np.zeros(position_space.shape)
    elif mode == 1:
        # Two-dimensional regular grid with checkerboard mask
        position_space = ift.RGSpace([128, 128])
        mask = make_checkerboard_mask(position_space)
    else:
        # Sphere with half of its pixels randomly masked
        position_space = ift.HPSpace(128)
        mask = make_random_mask(position_space)

    # Specify harmonic space corresponding to signal
    harmonic_space = position_space.get_default_codomain()

    # Harmonic transform from harmonic space to position space
    HT = ift.HarmonicTransformOperator(harmonic_space, target=position_space)

    # Set prior correlation covariance with a power spectrum leading to
    # homogeneous and isotropic statistics
    def power_spectrum(k):
        return 100./(20. + k**3)

    # 1D spectral space on which the power spectrum is defined
    power_space = ift.PowerSpace(harmonic_space)

    # Mapping to (higher dimensional) harmonic space
    PD = ift.PowerDistributor(harmonic_space, power_space)

    # Apply the mapping
    prior_correlation_structure = PD(ift.PS_field(power_space, power_spectrum))

    # Insert the result into the diagonal of an harmonic space operator
    S = ift.makeOp(prior_correlation_structure, sampling_dtype=float)
    # S is the prior field covariance

    # Build instrument response consisting of a discretization, mask
    # and harmonic transformaion

    # Masking operator to model that parts of the field have not been observed
    mask = ift.Field.from_raw(position_space, mask)
    Mask = ift.MaskOperator(mask)

    # The response operator consists of
    # - a harmonic transform (to get to image space)
    # - the application of the mask
    # - the removal of geometric information
    # The removal of geometric information is included in the MaskOperator
    # it can also be implemented with a GeometryRemover
    # Operators can be composed either with parenthesis
    R = Mask(HT)
    # or with @
    R = Mask @ HT

    data_space = R.target

    # Set the noise covariance N
    noise = 5.
    N = ift.ScalingOperator(data_space, noise, float)

    # Create mock data
    MOCK_SIGNAL = S.draw_sample()
    MOCK_NOISE = N.draw_sample()
    data = R(MOCK_SIGNAL) + MOCK_NOISE

    # Build inverse propagator D and information source j
    D_inv = R.adjoint @ N.inverse @ R + S.inverse
    j = R.adjoint_times(N.inverse_times(data))
    # Make D_inv invertible (via Conjugate Gradient)
    IC = ift.GradientNormController(iteration_limit=500, tol_abs_gradnorm=1e-3)
    D = ift.InversionEnabler(D_inv, IC, approximation=S.inverse).inverse

    # Calculate WIENER FILTER solution
    m = D(j)

    # Plotting
    rg = isinstance(position_space, ift.RGSpace)
    plot = ift.Plot()
    filename = "getting_started_1_mode_{}.png".format(mode)
    if rg and len(position_space.shape) == 1:
        plot.add(
            [HT(MOCK_SIGNAL), Mask.adjoint(data),
             HT(m)],
            label=['Mock signal', 'Data', 'Reconstruction'],
            alpha=[1, .3, 1])
        plot.add(Mask.adjoint(Mask(HT(m - MOCK_SIGNAL))), title='Residuals')
        plot.output(nx=2, ny=1, xsize=10, ysize=4, name=filename)
    else:
        plot.add(HT(MOCK_SIGNAL), title='Mock Signal')
        plot.add(Mask.adjoint(data), title='Data')
        plot.add(HT(m), title='Reconstruction')
        plot.add(Mask.adjoint(Mask(HT(m) - HT(MOCK_SIGNAL))),
                 title='Residuals')
        plot.output(nx=2, ny=2, xsize=10, ysize=10, name=filename)
    print("Saved results as '{}'.".format(filename))


if __name__ == '__main__':
    main()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
# Authors: Reimar Leike, Philipp Arras, Philipp Frank
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

###############################################################################
# Variational Inference (VI)
#
# This script demonstrates how MGVI, GeoVI, MeanfieldVI and FullCovarianceVI
# work for an inference problem with only two real quantities of interest. This
# enables us to plot the posterior probability density as two-dimensional plot.
###############################################################################

import numpy as np
import pylab as plt
from matplotlib.colors import LogNorm

import nifty.cl as ift


def main():
    dom = ift.UnstructuredDomain(1)
    scale = 10

    a = ift.FieldAdapter(dom, 'a')
    b = ift.FieldAdapter(dom, 'b')
    lh = (a.adjoint @ a).scale(scale) + (b.adjoint @ b).scale(-1.35*2).exp()
    lh = ift.VariableCovarianceGaussianEnergy(dom, 'a', 'b', np.float64) @ lh
    icsamp = ift.AbsDeltaEnergyController(deltaE=0.1, iteration_limit=2)
    ham = ift.StandardHamiltonian(lh, icsamp)

    x_limits = [-8/scale, 8/scale]
    x_limits_scaled = [-8, 8]
    y_limits = [-4, 4]
    x = np.linspace(*x_limits, num=401)
    y = np.linspace(*y_limits, num=401)
    xx, yy = np.meshgrid(x, y, indexing='ij')

    def np_ham(x, y):
        prior = x**2 + y**2
        mean = x*scale
        lcov = 1.35*2*y
        lh = .5*(mean**2*np.exp(-lcov) + lcov)
        return lh + prior

    z = np.exp(-1*np_ham(xx, yy))
    plt.plot(y, np.sum(z, axis=0))
    plt.xlabel('y')
    plt.ylabel('unnormalized pdf')
    plt.title('Marginal density')
    plt.pause(2.)
    plt.close()
    plt.plot(x*scale, np.sum(z, axis=1))
    plt.xlabel('x')
    plt.ylabel('unnormalized pdf')
    plt.title('Marginal density')
    plt.pause(2.)
    plt.close()

    mapx = xx[z == np.max(z)]
    mapy = yy[z == np.max(z)]
    meanx = (xx*z).sum()/z.sum()
    meany = (yy*z).sum()/z.sum()

    n_samples = 100
    minimizer = ift.NewtonCG(
        ift.GradientNormController(iteration_limit=3, name='Mini'))
    IC = ift.StochasticAbsDeltaEnergyController(0.5, iteration_limit=20,
                                                name='advi')
    stochastic_minimizer_mf = ift.ADVIOptimizer(IC, eta=0.3)
    stochastic_minimizer_fc = ift.ADVIOptimizer(IC, eta=0.3)
    posmg = posgeo = posmf = posfc = ift.from_random(ham.domain, 'normal')
    fc = ift.FullCovarianceVI(posfc, ham, 10, False, initial_sig=0.01)
    mf = ift.MeanFieldVI(posmf, ham, 10, False, initial_sig=0.01)

    fig, axs = plt.subplots(2, 2, figsize=[12, 8])
    axs = axs.flatten()

    def update_plot(runs):
        for axx, (nn, kl) in zip(axs, runs):
            axx.clear()
            axx.imshow(z.T, origin='lower',  cmap='gist_earth_r',
                       norm=LogNorm(vmin=1e-3, vmax=np.max(z)),
                       extent=x_limits_scaled + y_limits)
            xs, ys = [], []
            if isinstance(kl, ift.SampledKLEnergyClass):
                samples = kl.samples.iterator()
            else:
                samples = (kl.draw_sample() for _ in range(n_samples))
            mx, my = 0., 0.
            for samp in samples:
                a = samp.asnumpy()['a']
                xs.append(a)
                mx += a
                b = samp.asnumpy()['b']
                ys.append(b)
                my += b
            mx /= n_samples
            my /= n_samples
            axx.scatter(np.array(xs)*scale, np.array(ys),
                        label=f'{nn} samples')
            axx.scatter(mx*scale, my, label=f'{nn} mean')
            axx.scatter(mapx*scale, mapy, label='MAP')
            axx.scatter(meanx*scale, meany, label='Posterior mean')
            axx.set_title(nn)
            axx.set_xlim(x_limits_scaled)
            axx.set_ylim(y_limits)
            axx.legend(loc='lower right')
        axs[0].xaxis.set_visible(False)
        axs[1].xaxis.set_visible(False)
        axs[1].yaxis.set_visible(False)
        axs[2].set_xlabel('x')
        axs[2].set_ylabel('y')
        axs[3].yaxis.set_visible(False)
        axs[3].set_xlabel('x')
        plt.tight_layout()
        plt.draw()
        plt.pause(2.)

    for ii in range(10):
        if ii % 2 == 0:
            # Resample GeoVI and MGVI
            mgkl = ift.SampledKLEnergy(posmg, ham, n_samples, None, False)
            mini_samp = ift.NewtonCG(
                    ift.AbsDeltaEnergyController(1E-8, iteration_limit=5))
            geokl = ift.SampledKLEnergy(posgeo, ham, n_samples, mini_samp, False)

            runs = (("MGVI", mgkl), ("GeoVI", geokl),
                    ("MeanfieldVI", mf), ("FullCovarianceVI", fc))
            update_plot(runs)

        mgkl, _ = minimizer(mgkl)
        geokl, _ = minimizer(geokl)
        mf.minimize(stochastic_minimizer_mf)
        fc.minimize(stochastic_minimizer_fc)
        posmg = mgkl.position
        posgeo = geokl.position
        runs = (("MGVI", mgkl), ("GeoVI", geokl),
                ("MeanfieldVI", mf), ("FullCovarianceVI", fc))
        update_plot(runs)
    ift.logger.info('Finished')
    # Uncomment the following line in order to leave the plots open
    # plt.show()


if __name__ == '__main__':
    main()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .. import random
from ..domain_tuple import DomainTuple
from ..linearization import Linearization
from ..minimization.energy import Energy
from ..multi_domain import MultiDomain
from ..sugar import from_random
from ..utilities import (allreduce_sum, get_MPI_params_from_comm, myassert,
                         shareRange)


class EnergyAdapter(Energy):
    """Helper class which provides the traditional Nifty Energy interface to
    Nifty operators with a scalar target domain.

    Parameters
    -----------
    position : :class:`nifty.cl.field.Field` or :class:`nifty.cl.multi_field.MultiField`
        The position where the minimization process is started.
    op : EnergyOperator
        The expression computing the energy from the input data.
    constants : list of strings
        The component names of the operator's input domain which are assumed
        to be constant during the minimization process.
        If the operator's input domain is not a MultiField, this must be empty.
        Default: [].
    want_metric : bool
        If True, the class will provide a `metric` property. This should only
        be enabled if it is required, because it will most likely consume
        additional resources. Default: False.
    nanisinf : bool
        If true, nan energies which can happen due to overflows in the forward
        model are interpreted as inf. Thereby, the code does not crash on
        these occaisions but rather the minimizer is told that the position it
        has tried is not sensible.
    """

    def __init__(self, position, op, constants=[], want_metric=False,
                 nanisinf=False):
        if len(constants) > 0:
            cstpos = position.extract_by_keys(constants)
            _, op = op.simplify_for_constant_input(cstpos)
            varkeys = set(op.domain.keys()) - set(constants)
            position = position.extract_by_keys(varkeys)
        super(EnergyAdapter, self).__init__(position)
        self._op = op
        self._want_metric = want_metric
        lin = Linearization.make_var(position, want_metric)
        tmp = self._op(lin)
        self._val = tmp.val.asnumpy()[()]
        self._grad = tmp.gradient
        self._metric = tmp._metric
        self._nanisinf = bool(nanisinf)
        if self._nanisinf and np.isnan(self._val):
            self._val = np.inf

    def at(self, position):
        return EnergyAdapter(position, self._op, want_metric=self._want_metric,
                             nanisinf=self._nanisinf)

    @property
    def value(self):
        return self._val

    @property
    def gradient(self):
        return self._grad

    @property
    def metric(self):
        return self._metric

    def apply_metric(self, x):
        return self._metric(x)


class StochasticEnergyAdapter(Energy):
    """Provide the energy interface for an energy operator where parts of the
    input are averaged instead of optimized.

    Specifically, a set of standard normal distributed samples are drawn for
    the input corresponding to `keys` and each sample is inserted partially
    into `op`. The resulting operators are then averaged.  The subdomain that
    is not sampled is left a stochastic average of an energy with the remaining
    subdomain being the DOFs that are considered to be optimization parameters.

    Notes
    -----
    `StochasticEnergyAdapter` should never be created using the constructor,
    but rather via the factory function :attr:`make`.
    """
    def __init__(self, position, op, keys, local_ops, n_samples, comm, nanisinf,
                 noise, _callingfrommake=False):
        if not _callingfrommake:
            raise NotImplementedError
        super(StochasticEnergyAdapter, self).__init__(position)
        for lop in local_ops:
            myassert(position.domain == lop.domain)
        self._comm = comm
        self._local_ops = local_ops
        self._n_samples = n_samples
        self._nanisinf = nanisinf
        lin = Linearization.make_var(position)
        v, g = [], []
        for lop in self._local_ops:
            tmp = lop(lin)
            v.append(tmp.val.asnumpy())
            g.append(tmp.gradient)
        self._val = allreduce_sum(v, self._comm)[()]/self._n_samples
        if np.isnan(self._val) and self._nanisinf:
            self._val = np.inf
        self._grad = allreduce_sum(g, self._comm)/self._n_samples
        self._noise = noise

        self._op = op
        self._keys = keys

    @property
    def value(self):
        return self._val

    @property
    def gradient(self):
        return self._grad

    def at(self, position):
        return StochasticEnergyAdapter(position, self._op, self._keys,
                    self._local_ops, self._n_samples, self._comm, self._nanisinf,
                    self._noise, _callingfrommake=True)

    def apply_metric(self, x):
        lin = Linearization.make_var(self.position, want_metric=True)
        res = []
        for op in self._local_ops:
            res.append(op(lin).metric(x))
        return allreduce_sum(res, self._comm)/self._n_samples

    @property
    def metric(self):
        from .kl_energies import _SelfAdjointOperatorWrapper
        return _SelfAdjointOperatorWrapper(self.position.domain,
                                           self.apply_metric)

    def resample_at(self, position):
        return StochasticEnergyAdapter.make(position, self._op, self._keys,
                                            self._n_samples, self._comm)

    @staticmethod
    def make(position, op, sampling_keys, n_samples, mirror_samples,
             comm=None, nanisinf=False):
        """Factory function for StochasticEnergyAdapter.

        Parameters
        ----------
        position : :class:`nifty.cl.multi_field.MultiField`
            Values of the optimization parameters
        op : Operator
            The objective function of the optimization problem. Must have a
            scalar target. The domain must be a `MultiDomain` with its keys
            being the union of `sampling_keys` and `position.domain.keys()`.
        sampling_keys : iterable of String
            The keys of the subdomain over which the stochastic average of `op`
            should be performed.
        n_samples : int
            Number of samples used for the stochastic estimate.
        mirror_samples : boolean
            Whether the negative of the drawn samples are also used, as they are
            equally legitimate samples. If true, the number of used samples
            doubles.
        comm : MPI communicator or None
            If not None, samples will be distributed as evenly as possible
            across this communicator. If `mirror_samples` is set, then a sample
            and its mirror image will always reside on the same task.
        nanisinf : bool
            If true, nan energies, which can occur due to overflows in the
            forward model, are interpreted as inf which can be interpreted by
            optimizers.
        """
        myassert(op.target == DomainTuple.scalar_domain())
        samdom = {}
        if not isinstance(n_samples, int):
            raise TypeError
        for k in sampling_keys:
            if (k in position.domain.keys()) or (k not in op.domain.keys()):
                raise ValueError
            samdom[k] = op.domain[k]
        samdom = MultiDomain.make(samdom)
        noise = []
        sseq = random.spawn_sseq(n_samples)

        ntask, rank, _ = get_MPI_params_from_comm(comm)
        for i in range(*shareRange(n_samples, ntask, rank)):
            with random.Context(sseq[i]):
                rnd = from_random(samdom)
                noise.append(rnd)
                if mirror_samples:
                    noise.append(-rnd)
        local_ops = []
        for nn in noise:
            _, tmp = op.simplify_for_constant_input(nn)
            myassert(tmp.domain == position.domain)
            local_ops.append(tmp)
        n_samples = 2*n_samples if mirror_samples else n_samples
        return StochasticEnergyAdapter(position, op, sampling_keys, local_ops,
                              n_samples, comm, nanisinf, noise, _callingfrommake=True)

    def samples(self):
        return self._noise



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..minimization.energy_adapter import EnergyAdapter
from ..multi_field import MultiField
from ..operators.energy_operators import (InverseGammaEnergy,
                                          StandardHamiltonian)
from ..operators.scaling_operator import ScalingOperator
from ..operators.simple_linear_operators import ducktape


def make_adjust_variances_hamiltonian(a,
                                      xi,
                                      position,
                                      samples=[],
                                      scaling=None,
                                      ic_samp=None):
    """Creates a Hamiltonian for constant likelihood optimizations.

    Constructs a Hamiltonian to solve constant likelihood optimizations of the
    form phi = a * xi under the constraint that phi remains constant.

    xi is desired to be a Gaussian white Field, thus variations that are
    more easily represented by a should be absorbed in a.

    Parameters
    ----------
    a : Operator
        Gives the amplitude when evaluated at position.
    xi : Operator
        Field Adapter selecting a part of position.
        xi is desired to be a Gaussian white Field.
    position : :class:`nifty.cl.field.Field`, :class:`nifty.cl.multi_field.MultiField`
        Contains the initial values for the operators a and xi, to be adjusted
    samples : :class:`nifty.cl.field.Field`, :class:`nifty.cl.multi_field.MultiField`
        Residual samples of position.
    scaling : Float
        Optional rescaling of the Likelihood.
    ic_samp : Controller
        Iteration Controller for Hamiltonian.

    Returns
    -------
    :class:`nifty.cl.operators.energy_operators.StandardHamiltonian`
        A Hamiltonian that can be used for further minimization.
    """

    d = a*xi
    d = (d.conjugate()*d).real
    n = len(samples)
    if n > 0:
        d_eval = 0.
        for i in range(n):
            d_eval = d_eval + d.force(position + samples[i])
        d_eval = d_eval/n
    else:
        d_eval = d.force(position)

    x = (a.conjugate()*a).real
    if scaling is not None:
        x = ScalingOperator(x.target, scaling)(x)

    return StandardHamiltonian(InverseGammaEnergy(d_eval/2.)(x),
                               ic_samp=ic_samp)


def do_adjust_variances(position, A, minimizer, xi_key='xi', samples=[]):
    """Adjusts the variance of xi_key to be represented by amplitude_operator.

    Solves a constant likelihood optimization of the
    form phi = A * position[xi_key] under the constraint that phi remains
    constant.

    The field indexed by xi_key is desired to be a Gaussian white Field,
    thus variations that are more easily represented by A will be absorbed in
    A.

    Parameters
    ----------
    position : :class:`nifty.cl.field.Field`, :class:`nifty.cl.multi_field.MultiField`
        Contains the initial values for amplitude_operator and the key xi_key,
        to be adjusted.
    A : Operator
        Gives the amplitude when evaluated at position.
    minimizer : Minimizer
        Used to solve the optimization problem.
    xi_key : String
        Key of the Field containing undesired variations. This Field is
        contained in position.
    samples : :class:`nifty.cl.field.Field`, :class:`nifty.cl.multi_field.MultiField`, optional
        Residual samples of position. If samples are supplied then phi remains
        only approximately constant. Default: [].

    Returns
    -------
    MultiField
        The new position after variances have been adjusted.
    """
    xi = ducktape(None, position.domain, xi_key)
    ham = make_adjust_variances_hamiltonian(A, xi, position, samples=samples)

    # Minimize
    e = EnergyAdapter(
        position.extract(A.domain), ham, constants=[], want_metric=True)
    e, _ = minimizer(e)

    # Update position
    s_h_old = (A*xi).force(position)

    position = position.to_dict()
    position[xi_key] = s_h_old/A(e.position)
    position = MultiField.from_dict(position)
    return MultiField.union([position, e.position])



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2021 Max-Planck-Society
# Copyright(C) 2022-2025 Philipp Arras
# Copyright(C) 2025 LambdaFields GmbH
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import cProfile
import io
import pstats
import sys
from time import time
from warnings import warn

import numpy as np

from . import pointwise, utilities
from .any_array import AnyArray
from .domain_tuple import DomainTuple
from .domains.power_space import PowerSpace
from .field import Field
from .logger import logger
from .multi_domain import MultiDomain
from .multi_field import MultiField
from .operators.block_diagonal_operator import BlockDiagonalOperator
from .operators.diagonal_operator import DiagonalOperator
from .operators.distributors import PowerDistributor
from .operators.operator import Operator
from .operators.scaling_operator import ScalingOperator
from .operators.selection_operators import SliceOperator
from .plot import Plot, plottable2D

__all__ = ['PS_field', 'power_analyze', 'create_power_operator',
           'density_estimator', 'create_harmonic_smoothing_operator',
           'from_random', 'full', 'makeField', 'is_fieldlike',
           'is_linearization', 'is_operator', 'makeDomain', 'is_likelihood_energy',
           'get_signal_variance', 'makeOp', 'domain_union',
           'get_default_codomain', 'single_plot', 'exec_time',
           'recursive_operator_tree_profile', 'full_recursive_operator_tree_profile',
           'calculate_position', 'plot_priorsamples'] + list(pointwise.ptw_dict.keys())


def PS_field(pspace, function, device_id=-1):
    """Convenience function sampling a power spectrum

    Parameters
    ----------
    pspace : PowerSpace
        space at whose `k_lengths` the power spectrum function is evaluated
    function : function taking and returning a numpy.ndarray(float)
        the power spectrum function
    device_id : int

    Returns
    -------
    :class:`nifty.cl.field.Field`
        A field defined on (pspace,) containing the computed function values
    """
    if not isinstance(pspace, PowerSpace):
        raise TypeError
    data = function(pspace.k_lengths)
    return Field(DomainTuple.make(pspace), AnyArray(data).at(device_id))


def get_signal_variance(spec, space):
    """
    Computes how much a field with a given power spectrum will vary in space

    This is a small helper function that computes the expected variance
    of a harmonically transformed sample of this power spectrum.

    Parameters
    ---------
    spec: method
        a method that takes one k-value and returns the power spectrum at that
        location
    space: PowerSpace or any harmonic Domain
        If this function is given a harmonic domain, it creates the naturally
        binned PowerSpace to that domain.
        The field, for which the signal variance is then computed, is assumed
        to have this PowerSpace as naturally binned PowerSpace
    """
    if space.harmonic:
        space = PowerSpace(space)
    if not isinstance(space, PowerSpace):
        raise ValueError(
            "space must be either a harmonic space or Power space.")
    field = PS_field(space, spec)
    dist = PowerDistributor(space.harmonic_partner, space)
    k_field = dist(field)
    return k_field.weight(2).s_sum()


def _single_power_analyze(field, idx, binbounds):
    power_domain = PowerSpace(field.domain[idx], binbounds)
    pd = PowerDistributor(field.domain, power_domain, idx)
    return pd.adjoint_times(field.weight(1)).weight(-1)  # divides by bin size


# MR FIXME: this function is not well suited for analyzing more than one
# subdomain at once, because it allows only one set of binbounds.
def power_analyze(field, spaces=None, binbounds=None,
                  keep_phase_information=False):
    """Computes the power spectrum for a subspace of `field`.

    Creates a PowerSpace for the space addressed by `spaces` with the
    given binning and computes the power spectrum as a
    :class:`~nifty.cl.field.Field` over this PowerSpace. This can only
    be done if the subspace to be analyzed is a harmonic space. The
    resulting field has the same units as the square of the initial
    field.

    Parameters
    ----------
    field : :class:`nifty.cl.field.Field`
        The field to be analyzed
    spaces : None or int or tuple of int, optional
        The indices of subdomains for which the power spectrum shall be
        computed.
        If None, all subdomains will be converted.
        (default : None).
    binbounds : None or array-like, optional
        Inner bounds of the bins (default : None).
        if binbounds is None : bins are inferred.
    keep_phase_information : bool, optional
        If False, return a real-valued result containing the power spectrum
        of `field`.
        If True, return a complex-valued result whose real component
        contains the power spectrum computed from the real part of `field`,
        and whose imaginary component contains the power
        spectrum computed from the imaginary part of `field`.
        The absolute value of this result should be identical to the output
        of power_analyze with keep_phase_information=False.
        (default : False).

    Returns
    -------
    :class:`nifty.cl.field.Field`
        The output object. Its domain is a PowerSpace and it contains
        the power spectrum of `field`.
    """

    for sp in field.domain:
        if not sp.harmonic and not isinstance(sp, PowerSpace):
            logger.warning("WARNING: Field has a space in `domain` which is "
                           "neither harmonic nor a PowerSpace.")

    spaces = utilities.parse_spaces(spaces, len(field.domain))

    if len(spaces) == 0:
        raise ValueError("No space for analysis specified.")

    field_real = not utilities.iscomplextype(field.dtype)
    if (not field_real) and keep_phase_information:
        raise ValueError("cannot keep phase from real-valued input Field")

    if keep_phase_information:
        parts = [field.real*field.real, field.imag*field.imag]
    else:
        if field_real:
            parts = [field**2]
        else:
            parts = [field.real*field.real + field.imag*field.imag]

    for space_index in spaces:
        parts = [_single_power_analyze(part, space_index, binbounds)
                 for part in parts]

    return parts[0] + 1j*parts[1] if keep_phase_information else parts[0]


def _create_power_field(domain, power_spectrum):
    if not callable(power_spectrum):  # we have a Field defined on a PowerSpace
        if not isinstance(power_spectrum, Field):
            raise TypeError("Field object expected")
        if len(power_spectrum.domain) != 1:
            raise ValueError("exactly one domain required")
        if not isinstance(power_spectrum.domain[0], PowerSpace):
            raise TypeError("PowerSpace required")
        power_domain = power_spectrum.domain[0]
        fp = power_spectrum
    else:
        power_domain = PowerSpace(domain)
        fp = PS_field(power_domain, power_spectrum)

    return PowerDistributor(domain, power_domain)(fp)


def create_power_operator(domain, power_spectrum, space=None,
                          sampling_dtype=None):
    """Creates a diagonal operator with the given power spectrum.

    Constructs a diagonal operator that is defined on the specified domain.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
        Domain on which the power operator shall be defined.
    power_spectrum : callable or :class:`nifty.cl.field.Field`
        An object that contains the power spectrum as a function of k.
    space : int
        the domain index on which the power operator will work
    sampling_dtype : dtype or dict of dtype
        Specifies the dtype of the underlying Gaussian distribution.  Gaussian.
        If `sampling_dtype` is `None`, the operator cannot be used as a
        covariance, i.e. no samples can be drawn. Default: None.

    Returns
    -------
    DiagonalOperator
        An operator that implements the given power spectrum.
    """
    domain = DomainTuple.make(domain)
    space = utilities.infer_space(domain, space)
    field = _create_power_field(domain[space], power_spectrum)
    return DiagonalOperator(field, domain, space, sampling_dtype)


def density_estimator(domain, pad=1.0, cf_fluctuations=None,
                      cf_azm_uniform=None, prefix=""):
    from .domains.rg_space import RGSpace
    from .library.correlated_fields import CorrelatedFieldMaker
    from .library.special_distributions import UniformOperator

    cf_azm_uniform_sane_default = (1e-4, 1.0)
    cf_fluctuations_sane_default = {
        "scale": (0.5, 0.3),
        "cutoff": (4.0, 3.0),
        "loglogslope": (-6.0, 3.0)
    }

    domain = DomainTuple.make(domain)
    dom_scaling = 1. + np.broadcast_to(pad, (len(domain.axes), ))
    if cf_fluctuations is None:
        cf_fluctuations = cf_fluctuations_sane_default
    if cf_azm_uniform is None:
        cf_azm_uniform = cf_azm_uniform_sane_default

    domain_padded = []
    for d_scl, d in zip(dom_scaling, domain):
        if not isinstance(d, RGSpace) or d.harmonic:
            te = [f"unexpected domain encountered in `domain`: {domain}"]
            te += "expected a non-harmonic `RGSpace`"
            raise TypeError("\n".join(te))
        shape_padded = tuple((d_scl * np.array(d.shape)).astype(int))
        domain_padded.append(RGSpace(shape_padded, distances=d.distances))
    domain_padded = DomainTuple.make(domain_padded)

    # Set up the signal model
    azm_offset_mean = 0.0  # The zero-mode should be inferred only from the data
    cfmaker = CorrelatedFieldMaker(prefix)
    for i, d in enumerate(domain_padded):
        if isinstance(cf_fluctuations, (list, tuple)):
            cf_fl = cf_fluctuations[i]
        else:
            cf_fl = cf_fluctuations
        cfmaker.add_fluctuations_matern(d, **cf_fl, prefix=f"ax{i}")
    scalar_domain = DomainTuple.scalar_domain()
    uniform = UniformOperator(scalar_domain, *cf_azm_uniform)
    azm = uniform.ducktape("zeromode")
    cfmaker.set_amplitude_total_offset(azm_offset_mean, azm)
    correlated_field = cfmaker.finalize(0).clip(-10., 10.)
    normalized_amplitudes = cfmaker.get_normalized_amplitudes()

    domain_shape = tuple(d.shape for d in domain)
    slc = SliceOperator(correlated_field.target, domain_shape)
    signal = (slc @ correlated_field).exp()

    model_operators = {
        "correlated_field": correlated_field,
        "select_subset": slc,
        "amplitude_total_offset": azm,
        "normalized_amplitudes": normalized_amplitudes
    }

    return signal, model_operators


def create_harmonic_smoothing_operator(domain, space, sigma):
    """Creates an operator which smoothes a subspace of a harmonic domain.

    Parameters
    ----------
    domain: DomainTuple
        The total domain and target of the operator
    space : int
        the index of the subspace on which the operator acts.
        This must be a harmonic space
    sigma : float
        The sigma of the Gaussian smoothing kernel

    Returns
    -------
    DiagonalOperator
        The requested smoothing operator
    """
    kfunc = domain[space].get_fft_smoothing_kernel_function(sigma)
    return DiagonalOperator(kfunc(domain[space].get_k_length_array()), domain,
                            space)


def full(domain, val, device_id=-1):
    """Convenience function creating Fields/MultiFields with uniform values.

    Parameters
    ----------
    domain : Domainoid
        the intended domain of the output field
    val : scalar value
        the uniform value to be placed into all entries of the result

    Returns
    -------
    :class:`nifty.cl.field.Field` or:class:`nifty.cl.mulit_field.MultiField`
        The newly created uniform field
    """
    if isinstance(domain, (dict, MultiDomain)):
        return MultiField.full(domain, val, device_id)
    return Field.full(domain, val, device_id)


def from_random(domain, random_type='normal', dtype=np.float64, **kwargs):
    """Convenience function creating Fields/MultiFields with random values.

    Parameters
    ----------
    domain : Domainoid
        the intended domain of the output field
    random_type : 'pm1', 'normal', or 'uniform'
            The random distribution to use.
    dtype : type
        data type of the output field (e.g. numpy.float64)
        If the datatype is complex, each real an imaginary part have
        variance 1.
    **kwargs : additional parameters for the random distribution
        ('mean' and 'std' for 'normal', 'low' and 'high' for 'uniform')

    Returns
    -------
    :class:`nifty.cl.field.Field` or:class:`nifty.cl.mulit_field.MultiField`
        The newly created random field

    Notes
    -----
    When called with a multi-domain, the individual fields will be drawn in
    alphabetical order of the multi-domain's domain keys. As a consequence,
    renaming these keys may cause the multi-field to be filled with different
    random numbers, even for the same initial RNG state.
    """
    if isinstance(domain, (dict, MultiDomain)):
        return MultiField.from_random(domain, random_type, dtype, **kwargs)
    return Field.from_random(domain, random_type, dtype, **kwargs)


def makeField(domain, arr):
    """Convenience function creating Fields/MultiFields from Numpy arrays or
    dicts of Numpy arrays.

    Parameters
    ----------
    domain : Domainoid
        the intended domain of the output field
    arr : Numpy array if `domain` corresponds to a `DomainTuple`,
          dictionary of Numpy arrays if `domain` corresponds to a `MultiDomain`

    Returns
    -------
    :class:`nifty.cl.field.Field` or:class:`nifty.cl.mulit_field.MultiField`
        The newly created random field
    """
    if isinstance(domain, (dict, MultiDomain)):
        if not isinstance(arr, dict):
            raise TypeError("If `domain` is an instance of `MultiDomain`, `arr` must be a dict of Numpy arrays.")
        return MultiField.from_raw(domain, arr)
    if np.isscalar(arr):
        domain = makeDomain(domain)
        arr = np.broadcast_to(arr, domain.shape)
    return Field.from_raw(domain, arr)


def makeDomain(domain):
    """Convenience function creating DomainTuples/MultiDomains Domainoids.

    Parameters
    ----------
    domain : Domainoid (can be DomainTuple, MultiDomain, dict, Domain or list of Domains)
        the description of the requested (multi-)domain

    Returns
    -------
    DomainTuple or MultiDomain
        The newly created domain object
    """
    if isinstance(domain, (MultiDomain, dict)):
        return MultiDomain.make(domain)
    return DomainTuple.make(domain)


def makeOp(inp, dom=None, sampling_dtype=None):
    """Converts a Field or MultiField to a diagonal operator.

    Parameters
    ----------
    inp : None, :class:`nifty.cl.field.Field` or :class:`nifty.cl.multi_field.MultiField`
        - if None, None is returned.
        - if Field on scalar-domain, a ScalingOperator with the coefficient
            given by the Field is returned.
        - if Field, a DiagonalOperator with the coefficients given by this
            Field is returned.
        - if MultiField, a BlockDiagonalOperator with entries given by this
            MultiField is returned.

    dom : DomainTuple or MultiDomain
        if `inp` is a scalar, this is used as the operator's domain

    sampling_dtype : dtype or dict of dtypes
        If `inp` shall represent the diagonal covariance of a Gaussian
        probabilty distribution, `sampling_dtype` specifies if it is real or
        complex Gaussian. If `sampling_dtype` is `None`, the operator cannot be
        used as a covariance, i.e. no samples can be drawn. Default: None.

    Notes
    -----
    No volume factors are applied.
    """
    if inp is None:
        return None
    if np.isscalar(inp):
        if not isinstance(dom, (DomainTuple, MultiDomain)):
            raise TypeError("need proper `dom` argument")
        return ScalingOperator(dom, inp, sampling_dtype=sampling_dtype)
    if dom is not None:
        utilities.check_object_identity(dom, inp.domain)
    if inp.domain is DomainTuple.scalar_domain():
        return ScalingOperator(inp.domain, inp.val, sampling_dtype=sampling_dtype)
    if isinstance(inp, Field):
        return DiagonalOperator(inp, sampling_dtype=sampling_dtype)
    if isinstance(inp, MultiField):
        dct = {}
        for key, val in inp.items():
            if isinstance(sampling_dtype, dict):
                sdt = sampling_dtype[key]
            else:
                sdt = sampling_dtype
            dct[key] = makeOp(val, sampling_dtype=sdt)
        return BlockDiagonalOperator(inp.domain, dct)
    raise NotImplementedError


def domain_union(domains):
    """Computes the union of multiple DomainTuples/MultiDomains.

    Parameters
    ----------
    domains : list of DomainTuple or MultiDomain
        - if DomainTuple, all entries must be equal
        - if MultiDomain, there must not be any conflicting components
    """
    if isinstance(domains[0], DomainTuple):
        for dom in domains[1:]:
            utilities.check_object_identity(dom, domains[0])
        return domains[0]
    return MultiDomain.union(domains)


# Pointwise functions

_current_module = sys.modules[__name__]

for f in pointwise.ptw_dict.keys():
    def func(f):
        def func2(x, *args, **kwargs):
           return x.ptw(f, *args, **kwargs)
        return func2
    setattr(_current_module, f, func(f))


def get_default_codomain(domainoid, space=None):
    """For `RGSpace`, returns the harmonic partner domain.
    For `DomainTuple`, returns a copy of the object in which the domain
    indexed by `space` is substituted by its harmonic partner domain.
    In this case, if `space` is None, it is set to 0 if the `DomainTuple`
    contains exactly one domain.

    Parameters
    ----------
    domain: `RGSpace` or `DomainTuple`
        Domain for which to constuct the default harmonic partner
    space: int
        Optional index of the subdomain to be replaced by its default
        codomain. `domain[space]` must be of class `RGSpace`.
    """
    from .domains.gl_space import GLSpace
    from .domains.hp_space import HPSpace
    from .domains.lm_space import LMSpace
    from .domains.rg_space import RGSpace
    if isinstance(domainoid, RGSpace):
        return domainoid.get_default_codomain()
    if not isinstance(domainoid, DomainTuple):
        raise TypeError(
            'Works only on RGSpaces and DomainTuples containing those')
    space = utilities.infer_space(domainoid, space)
    if not isinstance(domainoid[space], (RGSpace, HPSpace, GLSpace, LMSpace)):
        raise TypeError("can only codomain structrued spaces")
    ret = [dom for dom in domainoid]
    ret[space] = domainoid[space].get_default_codomain()
    return DomainTuple.make(ret)


def single_plot(field, **kwargs):
    """Creates a single plot using `Plot`.
    Keyword arguments are passed to both `Plot.add` and `Plot.output`.
    """
    p = Plot()
    p.add(field, **kwargs)
    if 'title' in kwargs:
        del(kwargs['title'])
    p.output(**kwargs)


def plot_priorsamples(op, n_samples=5, common_colorbar=True, **kwargs):
    """Create a number of prior sample plots using `Plot`

    Parameters
    ----------
    op:
        Operator that mapping from standard Gaussian with covariance 1 to the prior distribution

    n_samples: int
        Number of prior samples for plotting

    Note
    ----
    Keyword arguments are passed to both `Plot.add` and `Plot.output`.
    """
    p = Plot()
    samples = list(op(from_random(op.domain)) for _ in range(n_samples))
    if common_colorbar:
        vmin = min(np.min(samples[i].asnumpy()) for i in range(n_samples))
        vmax = max(np.max(samples[i].asnumpy()) for i in range(n_samples))
    else:
        vmin = vmax = None
    if plottable2D(samples[0]):
        for i in range(n_samples):
            p.add(samples[i], vmin=vmin, vmax=vmax, **kwargs)
            if 'title' in kwargs:
                del(kwargs['title'])
    else:
        p.add(samples, **kwargs)
    p.output(**kwargs)


def calculate_position(operator, output):
    """Finds approximate preimage of an operator for a given output."""
    from .minimization.descent_minimizers import NewtonCG
    from .minimization.iteration_controllers import GradientNormController
    from .minimization.kl_energies import SampledKLEnergy
    from .operators.energy_operators import GaussianEnergy, StandardHamiltonian

    if not isinstance(operator, Operator):
        raise TypeError
    if output.domain != operator.target:
        raise TypeError
    if isinstance(output, MultiField):
        cov = 1e-3*max([np.max(np.abs(vv)) for vv in output.asnumpy().values()])**2
        invcov = ScalingOperator(output.domain, cov).inverse
        dtype = list(set([ff.dtype for ff in output.values()]))
        if len(dtype) != 1:
            raise ValueError('Only MultiFields with one dtype supported.')
        dtype = dtype[0]
    else:
        cov = 1e-3*np.max(np.abs(output.asnumpy()))**2
        dtype = output.dtype
    invcov = ScalingOperator(output.domain, cov, output.dtype).inverse
    d = output + invcov.draw_sample(from_inverse=True)
    lh = GaussianEnergy(d, invcov) @ operator
    H = StandardHamiltonian(
        lh, ic_samp=GradientNormController(iteration_limit=200))
    pos = 0.1*from_random(operator.domain)
    minimizer = NewtonCG(GradientNormController(iteration_limit=10, name='findpos'))
    for ii in range(3):
        logger.info(f'Start iteration {ii+1}/3')
        kl = SampledKLEnergy(pos, H, 3, None)
        kl, _ = minimizer(kl)
        pos = kl.position
    return pos


def is_likelihood_energy(obj):
    """Checks if object behaves like a likelihood energy.
    """
    return isinstance(obj, Operator) and obj.get_transformation() is not None


def is_operator(obj):
    """Checks if object is operator-like.

    Note
    ----
    A simple `isinstance(obj, ift.Operator)` does not give the expected result
    because, e.g., :class:`~nifty.cl.field.Field` inherits from
    :class:`~nifty.cl.operators.operator.Operator`.
    """
    return isinstance(obj, Operator) and obj.val is None


def is_linearization(obj):
    """Checks if object is linearization-like."""
    return isinstance(obj, Operator) and obj.jac is not None


def is_fieldlike(obj):
    """Checks if object is field-like.

    Note
    ----
    A simple `isinstance(obj, ift.Field)` does not give the expected result
    because users might have implemented another class which behaves field-like
    but is not an instance of :class:`~nifty.cl.field.Field`. Also note that
    instances of :class:`~nifty.cl.linearization.Linearization` behave
    field-like.
    """
    return isinstance(obj, Operator) and obj.val is not None


# -------------------------------------------------------------------------------
# Profiling Utilities
# -------------------------------------------------------------------------------
def exec_time(obj, want_metric=True, verbose=False, domain_dtype=np.float64, ntries=1,
              device_id=-1, dump_prefix=None):
    """Times the execution time of an operator or an energy.

    Parameters
    ----------

    obj : Operator or Energy
        Operator or Energy that shall be profiled.
    want_metric : bool, optional
        Determine if Operator shall be called with `want_metric=True`. Only
        applicable for EnergyOperators. Default: True.
    verbose : bool, optional
        If True, more profiling information is printed. Default: False.
    domain_dtype : dtype or dict of dtype

    ntries : int
        Number of times the operator shall be called. Default: 1.
    device_id : int or dict
        Device id of the input. If it is an `int`, all `Field`s of the
        `MultiField` reside on the same device. If it is a `dict`, the `Field`
        associated with a `key` of the `dict` will the reside on the decvice as
        indicated by the `dict`.
    dump_prefix : str or None
        If not None, profile files are written to disk with the respective
        prefix as defined by `dump_prefix` that, e.g., can be analyzed further
        with other tools. Default: None.
    """
    from .linearization import Linearization
    from .minimization.energy import Energy
    if device_id > -1:
        import cupy
        synchronize = cupy.cuda.Device(device_id).synchronize
    else:
        synchronize = lambda : None

    timing_results = {}

    def _profile_func(func, inp, what, save_key):
        # Warmup
        for _ in range(3):
            res = func(inp)
        # Profiling
        t0 = time()
        with cProfile.Profile() as pr:
            for _ in range(ntries):
                res = func(inp)
                synchronize()
        # TODO: Report both median and mean
        duration = (time() - t0)/ntries
        logger.info(f'{what}: {duration*1000:>8.3f} ms')
        if verbose:
            s = io.StringIO()
            pstats.Stats(pr, stream=s).sort_stats(pstats.SortKey.TIME).print_stats(5)
            logger.info(s.getvalue())
        if dump_prefix is not None:
            pr.dump_stats(f"{dump_prefix}_{save_key}.prof")
        timing_results[save_key] = duration
        return res

    def _profile_get_attr(obj, attr, what, save_key):
        return _profile_func(lambda x: getattr(obj, x), attr, what, save_key)

    if isinstance(obj, Energy):
        newpos = 0.99*obj.position
        _profile_func(lambda x: x.at(newpos), obj, "Energy.at()\t\t\t\t", "energy.at")
        _profile_get_attr(obj, "value", "Energy.value\t\t\t\t", "value")
        _profile_get_attr(obj, "gradient", "Energy.gradient\t\t\t\t", "gradient")
        _profile_get_attr(obj, "metric", "Energy.metric\t\t\t\t", "metric")
        if obj.metric is not None:
            _profile_func(lambda x: x.apply_metric(x.position), obj, "Energy.apply_metric\t\t\t", "apply_metric")
            _profile_func(lambda x: x.metric(x.position), obj, "Energy.metric(position)\t\t\t", "metric()")

    elif isinstance(obj, Operator):
        want_metric = bool(want_metric)

        pos = from_random(obj.domain, 'normal', dtype=domain_dtype, device_id=device_id)
        lin = Linearization.make_var(pos, want_metric=want_metric)
        _profile_func(lambda x: x(pos), obj, "Operator call with field\t\t", "apply")
        res = _profile_func(lambda x: x(lin), obj, "Operator call with linearization\t", "apply_lin")
        _profile_func(lambda x: res.jac(x), pos, "Apply linearization\t\t\t", "jac")
        _profile_func(lambda x: res.jac.adjoint(x), res.val, "Apply linearization (adjoint)\t\t", "jac.adjoint")

        if obj.target is DomainTuple.scalar_domain():
            _profile_get_attr(res, "gradient", "Gradient evaluation\t\t\t", "gradient")
            if want_metric:
                _profile_func(lambda x: res.metric(x), pos, "Metric apply\t\t\t\t", "metric_apply")
    else:
        raise TypeError

    return timing_results


def operator_tree_traverse(func, op, loc, depth=np.inf):
    """Recursively traverse a operator tree and apply a function at each leaf

    Parameters
    ----------
    func : callable
        A function to apply at each operator node. Should take two arguments:
        the operator (`op`) and the associated location (`loc`), and return a
        result.
    op : Operator
        The operator to traverse.
    loc : Field or MultiField
        The input data on which the operator acts. Must match the domain of the
        operator.
    depth : int
        The maximum depth to which the tree should be traversed. A depth of 0
        stops recursion and treats the current operator as a leaf. Default: inf.

    Returns
    -------
    dict
        A dictionary with the following keys:
        - 'result': The output of `func(op, loc)`.
        - 'name': The class name of the operator.
        - 'leaves': A tuple of recursively collected results from child operators, or an
                    empty tuple if `depth` is 0 or the operator is treated as a leaf.
    """
    from .library.correlated_fields import _Amplitude, _AmplitudeMatern
    from .linearization import Linearization
    from .operators.chain_operator import ChainOperator
    from .operators.energy_operators import (StandardHamiltonian,
                                             _LikelihoodChain)
    from .operators.operator import _OpChain, _OpProd, _OpSum
    from .operators.operator_adapter import OperatorAdapter
    from .operators.sampling_enabler import SamplingEnabler
    from .operators.sandwich_operator import SandwichOperator
    from .operators.sum_operator import SumOperator

    if op.domain is not loc.domain:
        raise ValueError(f"Domain mismatch. Got op.domain:\n"
                         f"{op.domain}\n\nGot loc.domain:\n{loc.domain}")
    if not isinstance(op, OperatorAdapter):
        name = type(op).__name__
    else:
        mode = [None, "adjoint", "inverse", "adjoint inverse"][op._trafo]
        name = f"{type(op._op).__name__}.{mode}"

    out = dict(result=func(op, loc), name=name)

    if depth == 0:
        out["leaves"] = tuple()
        return out
    depth = depth - 1

    def _reset_lin(x):
        if isinstance(x, Linearization):
            return Linearization.make_var(x.val)
        return x

    def _extract(x, domain):
        if isinstance(x, Linearization):
            return Linearization.make_var(x.val.extract(domain))
        return x.extract(domain)

    if isinstance(op, (_LikelihoodChain, _Amplitude, _AmplitudeMatern)):
        out["leaves"] = (operator_tree_traverse(func, op._op, loc, depth),)

    elif isinstance(op, StandardHamiltonian):
        out["leaves"] = tuple(operator_tree_traverse(func, oo, loc, depth)
                              for oo in [op._lh, op._prior])

    elif isinstance(op, SamplingEnabler):
        out["leaves"] = tuple(operator_tree_traverse(func, oo, loc, depth)
                              for oo in [op._likelihood, op._prior])

    elif isinstance(op, SandwichOperator):
        bun = operator_tree_traverse(func, op._bun, loc, depth)
        bun_adjoint = operator_tree_traverse(func, op._bun.adjoint,
                                             _reset_lin(op._bun(loc)), depth)
        cheese = operator_tree_traverse(func, op._cheese,
                                        _reset_lin(op._bun(loc)), depth)
        out["leaves"] = (bun_adjoint, cheese, bun)

    elif isinstance(op, (_OpChain, ChainOperator)):
        # Compute intermediate locations of operator chain
        locs = [loc]
        for oo in reversed(op._ops):
            locs.append(_reset_lin(oo(locs[-1])))
        locs.pop()  # Final result is not a input
        assert len(locs) == len(op._ops)
        locs = list(reversed(locs))
        for oo, ll in zip(op._ops, locs):
            assert oo.domain is ll.domain

        # Actually traverse
        out["leaves"] = tuple(operator_tree_traverse(func, oo, ll, depth)
                              for oo, ll in zip(op._ops, locs))

    elif isinstance(op, (_OpProd, _OpSum)):
        newloc = [_extract(loc, op._op1.domain),
                  _extract(loc, op._op2.domain)]
        out["leaves"] = tuple(operator_tree_traverse(func, oo, ll, depth) for oo, ll in
                              zip([op._op1, op._op2], newloc))

    elif isinstance(op, SumOperator):
        out["leaves"] = tuple(operator_tree_traverse(func, oo, _extract(loc, oo.domain), depth)
                              for oo in op._ops)

    elif isinstance(op, OperatorAdapter) and isinstance(op._op, SumOperator) and mode == "adjoint":
        out["leaves"] = tuple(operator_tree_traverse(func, oo.adjoint, _extract(loc, oo.target), depth)
                              for oo in op._op._ops)

    # Endpoints
    elif isinstance(op, OperatorAdapter):
        warn(f"{type(op)} treated as endpoint with children: {op._op}")
        out["leaves"] = tuple()

    else:
        warn(f"{type(op)} treated as endpoint")
        out["leaves"] = tuple()

    return out


def recursive_operator_tree_profile(op, loc, ntries, depth=np.inf):
    if isinstance(loc.domain, DomainTuple):
        device_id = loc.device_id
    else:
        device_id = set(loc.device_id.values())
        if len(device_id) > 1:
            raise NotImplementedError("`loc` is distributed across multiple devices")
        device_id = list(device_id)[0]
    assert isinstance(device_id, int) and device_id >= -1
    if op.domain is not loc.domain:
        raise ValueError("domain mismatch")

    if device_id > -1:
        import cupy
        synchronize = cupy.cuda.Device(device_id).synchronize
    else:
        synchronize = lambda : None

    def func(op, loc):
        # Warm up
        for _ in range(3):
            op(loc)
        synchronize()
        t0 = time()
        for _ in range(ntries):
            op(loc)
        synchronize()
        return f"{(time()-t0)*1000/ntries:.3f} ms"

    data = operator_tree_traverse(func, op, loc, depth)

    return _format_helper(data)


def full_recursive_operator_tree_profile(op, loc, ntries, depth=np.inf):
    from .linearization import Linearization

    out = dict()
    out["apply"] = recursive_operator_tree_profile(op, loc, ntries, depth)
    loc = Linearization.make_var(loc, want_metric=True)
    out["apply_lin"] = recursive_operator_tree_profile(op, loc, ntries, depth)
    res = op(loc)
    out["jac"] = recursive_operator_tree_profile(res.jac, loc, ntries, depth)
    out["jac.adjoint"] = recursive_operator_tree_profile(res.jac.adjoint, res.val, ntries, depth)
    if res.metric is not None:
        out["metric"] = recursive_operator_tree_profile(res.metric, loc, ntries, depth)
    return out


def _format_helper(data, _level=0):
    assert "result" in data
    assert "leaves" in data
    assert "name" in data

    s = f"{data['result']:>12}" + f" {_level:>3} " + _level*"  " + str(data["name"]) + "\n"
    if data["leaves"] != tuple():
        for newdata in data["leaves"]:
            s += _format_helper(newdata, _level=_level+1)
    return s



from .dof_space import DOFSpace
from .domain import Domain
from .gl_space import GLSpace
from .hp_space import HPSpace
from .lm_space import LMSpace
from .power_space import PowerSpace
from .rg_space import RGSpace
from .structured_domain import StructuredDomain
from .unstructured_domain import UnstructuredDomain



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .. import utilities
from ..domain_tuple import DomainTuple
from ..field import Field
from .linear_operator import LinearOperator


class ContractionOperator(LinearOperator):
    """A :class:`LinearOperator` which sums up fields into the direction of
    subspaces.

    This Operator sums up a field which is defined on a :class:`DomainTuple`
    to a :class:`DomainTuple` which is a subset of the former.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
    spaces : None, int or tuple of int
        The elements of "domain" which are contracted.
        If `None`, everything is contracted
    power : int, default=0
        If nonzero, the fields defined on self.domain are weighted with the
        specified power along the submdomains which are contracted.
    """

    def __init__(self, domain, spaces, power=0):
        self._domain = DomainTuple.make(domain)
        self._spaces = utilities.parse_spaces(spaces, len(self._domain))
        self._target = [
            dom for i, dom in enumerate(self._domain) if i not in self._spaces
        ]
        self._target = DomainTuple.make(self._target)
        self._power = power
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        dev0 = x.device_id
        if mode == self.ADJOINT_TIMES:
            ldat = x.val
            shp = []
            for i, dom in enumerate(self._domain):
                tmp = dom.shape
                shp += tmp if i not in self._spaces else (1,)*len(dom.shape)
            ldat = np.broadcast_to(ldat.reshape(shp), self._domain.shape)
            res = Field(self._domain, ldat)
            if self._power != 0:
                res = res.weight(self._power, spaces=self._spaces)
            assert res.device_id == dev0
        else:
            if self._power != 0:
                x = x.weight(self._power, spaces=self._spaces)
            assert x.device_id == dev0
            res = x.sum(self._spaces)
            assert res.device_id == dev0
            if np.isscalar(res):
                res = Field.scalar(res).at(dev0, check_fail=False)
            assert res.device_id == dev0
        return res


def IntegrationOperator(domain, spaces):
    """A :class:`LinearOperator` which integrates fields into the direction
    of subspaces.

    This Operator integrates a field which is defined on a :class:`DomainTuple`
    to a :class:`DomainTuple` which is a subset of the former.

    Parameters
    ----------
    domain : Domain, tuple of Domain or DomainTuple
    spaces : None, int or tuple of int
        The elements of "domain" which are contracted.
        If `None`, everything is contracted
    """
    return ContractionOperator(domain, spaces, 1)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from .. import utilities
from ..any_array import AnyArray
from ..domain_tuple import DomainTuple
from ..field import Field
from .endomorphic_operator import EndomorphicOperator


class MatrixProductOperator(EndomorphicOperator):
    """Endomorphic matrix multiplication with input field.

    This operator supports scipy.sparse matrices and numpy arrays
    as the matrix to be applied.

    For numpy array matrices, can apply the matrix over a subspace
    of the input.

    If the input arrays have more than one dimension, for
    scipy.sparse matrices the `flatten` keyword argument must be
    set to true. This means that the input field will be flattened
    before applying the matrix and reshaped to its original shape
    afterwards.

    Matrices are tested regarding their compatibility with the
    called for application method.

    Flattening and subspace application are mutually exclusive.

    Parameters
    ----------
    domain: Domain or DomainTuple
        Domain of the operator.
        If :class:`DomainTuple` it is assumed to have only one entry.
    matrix: scipy.sparse.spmatrix or numpy.ndarray
        Quadratic matrix of shape `(domain.shape, domain.shape)`
        (if `not flatten`) that supports `matrix.transpose()`.
        If it is not a numpy array, needs to be applicable to the val
        array of input fields by `matrix.dot()`.
    spaces: int or tuple of int, optional
        The subdomain(s) of "domain" which the operator acts on.
        If None, it acts on all elements.
        Only possible for numpy array matrices.
        If `len(domain) > 1` and `flatten=False`, this parameter is
        mandatory.
    flatten: boolean, optional
        Whether the input value array should be flattened before
        applying the matrix and reshaped to its original shape
        afterwards.
        Needed for scipy.sparse matrices if `len(domain) > 1`.
    """
    def __init__(self, domain, matrix, spaces=None, flatten=False):
        self._capability = self.TIMES | self.ADJOINT_TIMES
        self._domain = DomainTuple.make(domain)
        self._flatten = flatten

        mat_dim = len(matrix.shape)

        if mat_dim % 2 != 0 or \
           matrix.shape != (matrix.shape[:mat_dim//2] + matrix.shape[:mat_dim//2]):
            raise ValueError("Matrix must be quadratic.")
        appl_dim = mat_dim // 2  # matrix application space dimension

        # take shortcut for trivial case
        if spaces is not None:
            if len(self._domain.shape) == 1 and spaces == (0, ):
                spaces = None

        if spaces is None:
            self._spaces = None
            self._active_axes = utilities.my_sum(self._domain.axes)
            appl_space_shape = self._domain.shape
            if flatten:
                appl_space_shape = (utilities.my_product(appl_space_shape), )
        else:
            if flatten:
                raise ValueError(
                    "Cannot flatten input AND apply to a subspace")
            if not isinstance(matrix, np.ndarray):
                raise ValueError(
                    "Application to subspaces only supported for numpy array matrices."
                )
            self._spaces = utilities.parse_spaces(spaces, len(self._domain))
            appl_space_shape = []
            active_axes = []
            for space_idx in spaces:
                appl_space_shape += self._domain[space_idx].shape
                active_axes += self._domain.axes[space_idx]
            appl_space_shape = tuple(appl_space_shape)
            self._active_axes = tuple(active_axes)

            self._mat_last_n = tuple([-appl_dim + i for i in range(appl_dim)])
            self._mat_first_n = tuple(range(appl_dim))

        # Test if the matrix and the array it will be applied to fit
        if matrix.shape[:appl_dim] != appl_space_shape:
            raise ValueError(
                "Matrix and domain shapes are incompatible under the requested "
                + "application scheme.\n" +
                f"Matrix appl shape: {matrix.shape[:appl_dim]}, " +
                f"appl_space_shape: {appl_space_shape}.")

        self._mat = AnyArray(matrix)
        self._finalize_init()

    def _finalize_init(self):
        self._mat_tr = np.transpose(self._mat).conjugate()

    def _device_preparation(self, x, mode):
        self._mat = self._mat.at(x.device_id)
        self._finalize_init()

    def apply(self, x, mode):
        self._check_input(x, mode)
        self._device_preparation(x, mode)
        times = (mode == self.TIMES)
        m = self._mat if times else self._mat_tr
        if isinstance(m, AnyArray):
            print(times)
            assert m.device_id == x.device_id

        if self._spaces is None:
            if not self._flatten:
                res = np.dot(m, x.val)
            else:
                res = np.dot(m, x.val.flatten()).reshape(self._domain.shape)
            return Field(self._domain, res)

        mat_axes = self._mat_last_n if times else tuple(reversed(self._mat_last_n))
        move_axes = self._mat_first_n if times else tuple(reversed(self._mat_first_n))
        res = np.tensordot(m, x.val, axes=(mat_axes, self._active_axes))
        res = np.moveaxis(res, move_axes, self._active_axes)
        return Field(self._domain, res)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from .multi_field import MultiField
from .operators.endomorphic_operator import EndomorphicOperator
from .operators.operator import Operator
from .sugar import from_random, makeField


class StatCalculator:
    """Helper class to compute mean and variance of a set of inputs.

    Notes
    -----
    - The memory usage of this object is constant, i.e. it does not increase
      with the number of samples added.
    - The code computes the unbiased variance (which contains a `1./(n-1)`
      term for `n` samples).
    """
    def __init__(self):
        self._count = 0

    def add(self, value):
        """Adds a sample.

        Parameters
        ----------
        value: any type that supports multiplication by a scalar and
               element-wise addition/subtraction/multiplication.
        """
        self._count += 1
        if self._count == 1:
            self._mean = 1.*value
            self._M2 = 0.*value
        else:
            delta = value - self._mean
            self._mean = self.mean + delta*(1./self._count)
            delta2 = value - self._mean
            self._M2 = self._M2 + delta*delta2

    @property
    def mean(self):
        """
        value type : the mean of all samples added so far.
        """
        if self._count == 0:
            raise RuntimeError
        return 1.*self._mean

    @property
    def var(self):
        """
        value type : the unbiased variance of all samples added so far.
        """
        if self._count < 2:
            raise RuntimeError
        return self._M2 * (1./(self._count-1))


def probe_with_posterior_samples(op, post_op, nprobes, dtype):
    '''FIXME

    Parameters
    ----------
    op : EndomorphicOperator
        FIXME
    post_op : Operator
        FIXME
    nprobes : int
        Number of samples which shall be drawn.
    dtype :
        the data type of the samples

    Returns
    -------
    List of :class:`nifty.cl.field.Field`
        List of two fields: the mean and the variance.
    '''
    if not isinstance(op, EndomorphicOperator):
        raise TypeError
    if post_op is not None:
        if not isinstance(post_op, Operator):
            raise TypeError
        if post_op.domain is not op.target:
            raise ValueError
    sc = StatCalculator()
    for i in range(nprobes):
        if post_op is None:
            sc.add(op.draw_sample(from_inverse=True))
        else:
            sc.add(post_op(op.draw_sample(from_inverse=True)))

    if nprobes == 1:
        return sc.mean, None
    return sc.mean, sc.var


def probe_diagonal(op, nprobes, random_type="pm1"):
    '''Probes the diagonal of an endomorphic operator.

    The operator is called on a user-specified number of randomly generated
    input vectors :math:`v_i`, producing :math:`r_i`. The estimated diagonal
    is the mean of :math:`r_i^\\dagger v_i`.

    Parameters
    ----------
    op: EndomorphicOperator
        The operator to be probed.
    nprobes: int
        The number of probes to be used.
    random_type: str
        The kind of random number distribution to be used for the probing.
        The default value `pm1` causes the probing vector to be randomly
        filled with values of +1 and -1.

    Returns
    -------
    :class:`nifty.cl.field.Field`
        The estimated diagonal.
    '''
    sc = StatCalculator()
    for i in range(nprobes):
        x = from_random(op.domain, random_type)
        sc.add(op(x).conjugate()*x)
    return sc.mean


def approximation2endo(op, nsamples):
    sc = StatCalculator()
    for _ in range(nsamples):
        sc.add(op.draw_sample())
    approx = sc.var
    dct = approx.to_dict()
    for kk in dct:
        foo = dct[kk].asnumpy_rw()
        foo[foo == 0] = 1
        dct[kk] = makeField(dct[kk].domain, foo)
    return MultiField.from_dict(dct)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2020 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..field import Field
from .structured_domain import StructuredDomain


class LMSpace(StructuredDomain):
    """Represents a set of spherical harmonic coefficients.

    Its harmonic partner spaces are :class:`~nifty.cl.domains.hp_space.HPSpace`
    and :class:`~nifty.cl.domains.gl_space.GLSpace`.

    Parameters
    ----------
    lmax : int
        The maximum :math:`l` value of any spherical harmonic coefficient
        :math:`a_{lm}` that is represented by this object.
        Must be :math:`\\ge 0`.

    mmax : int, optional
        The maximum :math:`m` value of any spherical harmonic coefficient
        :math:`a_{lm}` that is represented by this object.
        If not supplied, it is set to `lmax`.
        Must be :math:`\\ge 0` and :math:`\\le` `lmax`.
    """

    _needed_for_hash = ["_lmax", "_mmax"]

    def __init__(self, lmax, mmax=None):
        self._lmax = int(lmax)
        if self._lmax < 0:
            raise ValueError("lmax must be >=0.")
        if mmax is None:
            mmax = self._lmax
        self._mmax = int(mmax)
        if self._mmax < 0 or self._mmax > self._lmax:
            raise ValueError("mmax must be >=0 and <=lmax.")

    def __repr__(self):
        return "LMSpace(lmax={}, mmax={})".format(self.lmax, self.mmax)

    @property
    def harmonic(self):
        return True

    @property
    def shape(self):
        return (self.size,)

    @property
    def size(self):
        l, m = self._lmax, self._mmax
        # the LMSpace consists of the full triangle (including -m's!),
        # minus two little triangles if mmax < lmax
        return (l+1)**2 - (l-m)*(l-m+1)

    @property
    def scalar_dvol(self):
        return 1.

    def get_k_length_array(self):
        lmax = self._lmax
        mmax = self._mmax
        ldist = np.empty((self.size,), dtype=np.float64)
        ldist[0:lmax+1] = np.arange(lmax+1, dtype=np.float64)
        tmp = np.empty((2*lmax+2), dtype=np.float64)
        tmp[0::2] = np.arange(lmax+1)
        tmp[1::2] = np.arange(lmax+1)
        idx = lmax+1
        for m in range(1, mmax+1):
            ldist[idx:idx+2*(lmax+1-m)] = tmp[2*m:]
            idx += 2*(lmax+1-m)
        return Field.from_raw(self, ldist)

    def get_unique_k_lengths(self):
        return np.arange(self.lmax+1, dtype=np.float64)

    @staticmethod
    def _kernel(x, sigma):
        # cf. "All-sky convolution for polarimetry experiments"
        # by Challinor et al.
        # https://arxiv.org/abs/astro-ph/0008228
        from ..sugar import exp
        return exp((x+1.) * x * (-0.5*sigma*sigma))

    def get_fft_smoothing_kernel_function(self, sigma):
        return lambda x: self._kernel(x, sigma)

    def get_conv_kernel_from_func(self, func):
        """Creates a convolution kernel defined by a function.

        Assumes the function to be radially symmetric, e.g. only dependant on
        theta in radians.

        Parameters
        ----------
        func: function
            This function needs to take exactly one argument, which is
            colatitude in radians, and return the kernel amplitude at that
            colatitude.
        """
        from ducc0.misc import GL_thetas

        from ..operators.harmonic_operators import HarmonicTransformOperator
        from .gl_space import GLSpace

        # define azimuthally symmetric spaces for kernel transform
        gl = GLSpace(self.lmax + 1, 1)
        lm0 = gl.get_default_codomain()
        theta = GL_thetas(gl.nlat)
        # evaluate the kernel function at the required thetas
        kernel_sphere = Field.from_raw(gl, func(theta))
        # normalize the kernel such that the integral over the sphere is 4pi
        kernel_sphere = kernel_sphere * (4 * np.pi / kernel_sphere.s_integrate())
        # compute the spherical harmonic coefficients of the kernel
        op = HarmonicTransformOperator(lm0, gl)
        kernel_lm = op.adjoint_times(kernel_sphere.weight(1)).asnumpy()
        # evaluate the k lengths of the harmonic space
        k_lengths = self.get_k_length_array().asnumpy().astype(np.int64)
        return Field.from_raw(self, kernel_lm[k_lengths])

    @property
    def lmax(self):
        """int : maximum allowed :math:`l`

        The maximum :math:`l` value of any spherical harmonic
        coefficient :math:`a_{lm}` that is represented in this domain.
        """
        return self._lmax

    @property
    def mmax(self):
        """int : maximum allowed :math:`m`

        The maximum :math:`m` value of any spherical harmonic
        coefficient :math:`a_{lm}` that is represented in this domain.
        """
        return self._mmax

    def get_default_codomain(self):
        """Returns a :class:`~nifty.cl.domains.gl_space.GLSpace` object, which is
        capable of storing an accurate representation of data residing on
        `self`.

        Returns
        -------
        GLSpace
            The partner domain
        """
        from ..domains.gl_space import GLSpace
        return GLSpace(self.lmax+1, self.mmax*2+1)

    def check_codomain(self, codomain):
        """Raises `TypeError` if `codomain` is not a matching partner domain
        for `self`.

        Notes
        -----
        This function only checks whether `codomain` is of type
        :class:`GLSpace` or :class:`HPSpace`.
        """
        from ..domains.gl_space import GLSpace
        from ..domains.hp_space import HPSpace
        if not isinstance(codomain, (GLSpace, HPSpace)):
            raise TypeError("codomain must be a GLSpace or HPSpace.")



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import sys
from functools import partial
from typing import Tuple

import jax
import numpy as np
from jax import numpy as jnp
from numpy.typing import NDArray

from ..logger import logger


def unique(ar, *, return_inverse=False, axis=-1, atol=1e-10, rtol=1e-5, _verbosity=0):
    """Find unique sub-arrays in `ar` along `axis` within a given tolerance.

    The algorithm is efficient if the number of approximately unique values is
    small compared to the overall size of `ar` along `axis`.
    """
    if not isinstance(axis, int):
        raise TypeError(f"`axis` needs to be of type `int`; got {type(axis)!r}")

    uniqs = None
    inverse = None
    if return_inverse:
        inverse = np.full(ar.shape[axis], -1, dtype=int)
    # Ensure positive axis required for identify the axis of reductions
    axis = np.arange(ar.ndim)[axis]
    ra = tuple(
        set(range(ar.ndim))
        - {
            axis,
        }
    )
    to_sort = np.ones(ar.shape[axis], dtype=bool)
    while np.sum(to_sort) != 0:
        i = np.nonzero(to_sort)[0][0]
        u = np.take(ar, (i,), axis=axis)
        if uniqs is None:
            uniqs = u
        else:
            uniqs = np.concatenate((uniqs, u), axis=axis)
        isclose = np.zeros(to_sort.shape, dtype=bool)
        # Set the `mode` to work around `ar` potentially being a JAX array and
        # not supporting NumPy's default `mode='raise'`
        a = np.take(ar, np.nonzero(to_sort)[0], axis=axis, mode=None)
        isclose[to_sort] = np.all(np.abs(u - a) <= (atol + rtol * np.abs(a)), axis=ra)
        to_sort &= ~isclose
        if return_inverse:
            assert inverse is not None
            inverse[isclose] = uniqs.shape[axis] - 1
        if _verbosity > 0:
            logger.info(f"to-sort: {np.sum(to_sort)}", file=sys.stderr)

    if return_inverse:
        assert np.all(inverse != -1)
        return uniqs, inverse
    return uniqs


def amend_unique(ar, el, *, axis=-1, atol=1e-10, rtol=1e-5) -> Tuple[NDArray, int]:
    """Amend the element `el` if it is unique up to the specified tolerance
    otherwise do nothing.
    """
    if not isinstance(axis, int):
        raise TypeError(f"`axis` needs to be of type `int`; got {type(axis)!r}")

    # Ensure positive axis required for identify the axis of reductions
    axis = np.arange(ar.ndim)[axis]
    ra = tuple(
        set(range(ar.ndim))
        - {
            axis,
        }
    )

    el = np.expand_dims(el, axis=axis)
    isclose = np.all(np.abs(ar - el) <= (atol + rtol * np.abs(el)), axis=ra)
    assert isclose.size == ar.shape[axis]
    if np.any(isclose):
        return ar, np.nonzero(isclose)[0][0]
    else:
        return np.concatenate((ar, el), axis=axis), ar.shape[axis]


@partial(jax.jit, static_argnames=("axis",))
def amend_unique_(ar, el, *, axis=-1, atol=1e-10, rtol=1e-5):
    if not isinstance(axis, int):
        raise TypeError(f"`axis` needs to be of type `int`; got {type(axis)!r}")
    PLC = -1 << 63 if jnp.array(0).dtype == jnp.int64 else -1 << 31

    # Ensure positive axis required for identify the axis of reductions
    axis = np.arange(ar.ndim)[axis]
    ra = tuple(
        set(range(ar.ndim))
        - {
            axis,
        }
    )

    el = jnp.expand_dims(el, axis=axis)
    isclose = jnp.all(jnp.abs(ar - el) <= (atol + rtol * jnp.abs(el)), axis=ra)

    # Find the first not-NaN location in the array at which to potentially
    # insert a new value
    n = jnp.nonzero(jnp.all(jnp.isnan(ar), axis=ra), size=1, fill_value=PLC)[0][0]

    # Replace NaN with NaN if the new element is close to any existing element,
    # else insert it at the first not-NaN location
    any_isclose = jnp.any(isclose)
    e = jnp.where(any_isclose, jnp.full_like(el, jnp.nan), el)
    ar = ar.at[(slice(None),) * axis + (n,)].set(jnp.squeeze(e, axis=axis))
    idx = jnp.nonzero(isclose, size=1, fill_value=PLC)[0][0]
    idx = jnp.where(any_isclose, idx, n)
    return ar, idx



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .. import config
from . import structured_kernel_interpolation
from .conjugate_gradient import cg, static_cg
from .correlated_field import CorrelatedFieldMaker
from .custom_map import lmap, smap
from .evi import (
    Samples,
    draw_linear_residual,
    draw_residual,
    nonlinearly_update_residual,
    wiener_filter_posterior,
)
from .evidence_lower_bound import estimate_evidence_lower_bound
from .extra import SamplingCartesianGridLOS
from .gauss_markov import (
    GaussMarkovProcess,
    IntegratedWienerProcess,
    OrnsteinUhlenbeckProcess,
    WienerProcess,
)
from .hmc import generate_hmc_acc_rej, generate_nuts_tree
from .hmc_oo import HMCChain, NUTSChain
from .likelihood import Likelihood, LikelihoodPartial
from .likelihood_impl import (
    Categorical,
    Gaussian,
    Poissonian,
    StudentT,
    VariableCovarianceGaussian,
    NDVariableCovarianceGaussian,
    VariableCovarianceStudentT,
)
from .logger import logger
from .minisanity import minisanity, reduced_residual_stats
from .misc import hvp, interpolate, wrap, wrap_left
from .model import Initializer, Model, VModel, WrappedCall
from .multi_grid import (
    Grid,
    HEALPixGrid,
    HPLogRGrid,
    HPBrokenLogRGrid,
    ICRField,
    LogGrid,
    BrokenLogGrid,
    MGrid,
    SimpleOpenGrid,
)
from .num import *
from .optimize import minimize, newton_cg, static_newton_cg, trust_ncg
from .optimize_kl import OptimizeVI, OptimizeVIState, optimize_kl
from .prior import (
    InvGammaPrior,
    LaplacePrior,
    LogNormalPrior,
    NormalPrior,
    UniformPrior,
)
from .refine.chart import CoordinateChart, HEALPixChart
from .refine.charted_field import ChartedField
from .refine.healpix_field import ChartedHPField
from .tree_math import *



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

import numpy as np

from ..domain_tuple import DomainTuple
from ..field import Field
from .linear_operator import LinearOperator


class ValueInserter(LinearOperator):
    """Inserts one value into a field which is zero otherwise.

    Parameters
    ----------
    target : Domain, tuple of Domain or DomainTuple
    index : iterable of int
        The index of the target into which the value of the domain shall be
        inserted.
    """

    def __init__(self, target, index):
        self._domain = DomainTuple.scalar_domain()
        self._target = DomainTuple.make(target)
        index = tuple(index)
        if not all([
                isinstance(n, int) and n >= 0 and n < self.target.shape[i]
                for i, n in enumerate(index)
        ]):
            raise TypeError
        if not len(index) == len(self.target.shape):
            raise ValueError
        self._index = index
        self._capability = self.TIMES | self.ADJOINT_TIMES
        # Check whether index is in bounds
        np.empty(self.target.shape)[self._index]

    def apply(self, x, mode):
        self._check_input(x, mode)
        x = x.val
        if mode == self.TIMES:
            res = np.zeros_like(x, shape=self.target.shape, dtype=x.dtype)
            res[self._index] = x
            return Field(self._tgt(mode), res)
        else:
            return Field.scalar(x[self._index], x.device_id)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import jax
from jax import numpy as jnp
from jax.tree_util import tree_map


def _check(v, cut=1e-16):
    return v > cut


@jax.custom_jvp
def _solve(A, X):
    assert A.ndim == 2
    assert (X.ndim == 1) or (X.ndim == 2)
    v, U = jnp.linalg.eigh(A)
    vi = jax.lax.select(_check(v), 1.0 / v, jnp.zeros_like(v))
    res = U.T @ X
    res *= vi[:, jnp.newaxis] if X.ndim == 2 else vi
    return U @ res


@_solve.defjvp
def _solve_jvp(primals, tangents):
    # Note: Makes use of `stable_inverse` directly to enable stable higher order
    # derivatives. This is a tradeoff against saving compute for first order.
    (A, X), (dA, dX) = primals, tangents
    res = _solve(A, X)
    return res, _solve(A, dX - dA @ res)


def solve(A, B, *, matrix_eqn=False, transposed=False):
    """
    Solves the linear system of equations

    .. math ::
        AX = B

    for X, where A is a symmetric positive definite matrix.
    Mapped over the leafs of the input trees and batched over leading axes.
    This implementation features a custom, more stable jvp rule.

    Parameters
    ----------
    A : tree-like structure of jnp.ndarray
        LHS of the equation
    B : tree-like structure of jnp.ndarray
        RHS of the equation
    matrix_eqn : bool
        Specifies whether X and B are matrices or Vectors (default: False)
    transposed: bool
        Solves the transposed system of equations

        .. math ::
            A^T X^T = B^T

        instead. (default: False)

    Returns
    -------
    out : tree-like structure of jnp.ndarray (like B)
        Solution X of the above equation

    """
    sig = "(m,m),(m,n)->(m,n)" if matrix_eqn else "(m,m),(m)->(m)"
    solve_leaf = jnp.vectorize(_solve, signature=sig)
    if transposed:
        A = tree_map(jnp.matrix_transpose, A)
        B = tree_map(jnp.matrix_transpose, B)
    res = tree_map(solve_leaf, A, B)
    if transposed:
        res = tree_map(jnp.matrix_transpose, res)
    return res


def _get_sqrt(v, U):
    vsq = jax.lax.select(_check(v), jnp.sqrt(v), jnp.zeros_like(v))
    return U @ (vsq[:, jnp.newaxis] * U.T)


@jax.custom_jvp
def _sqrtm(M):
    v, U = jnp.linalg.eigh(M)
    return _get_sqrt(v, U)


@_sqrtm.defjvp
def _sqrtm_jvp(M, dM):
    # Note: Only stable 1st derivative!
    M, dM = M[0], dM[0]
    v, U = jnp.linalg.eigh(M)

    dM = U.T @ dM @ U
    valid = _check(v)
    vsq = jnp.sqrt(jax.lax.select(valid, v, jnp.ones_like(v)))
    dres = jax.lax.select(
        valid[:, jnp.newaxis] & valid[jnp.newaxis, :],
        dM / (vsq[:, jnp.newaxis] + vsq[jnp.newaxis, :]),
        jnp.zeros_like(dM),
    )
    return _get_sqrt(v, U), U @ dres @ U.T


def sqrtm(M):
    """
    Computes the principal matrix squre root B of a symmetric square matrix M:

    .. math ::
        M = BB

    with B being positive semidefinite and symmetric.
    Mapped over the leafs of the input tree and batched over leading axes.
    This implementation features a custom, more stable jvp rule (only 1st derivative).

    Parameters
    ----------
    M : tree-like structure of jnp.ndarray
        The input matrix/matrices

    Returns
    -------
    B : tree-like structure of jnp.ndarray (like M)
        Principal matrix square root of M
    """
    sig = "(n,n)->(n,n)"
    return tree_map(jnp.vectorize(_sqrtm, signature=sig), M)


def _logm(M):
    v, U = jnp.linalg.eigh(M)
    vlog = jnp.log(v)
    return U @ (vlog[:, jnp.newaxis] * U.T)


def logm(M):
    """
    Computes the matrix logarithm of a symmetric square matrix.
    Input matrix must have strictly positive eigenvalues.
    Mapped over the leafs of the input tree and batched over leading axes.

    Parameters
    ----------
    M : tree-like structure of jnp.ndarray
        The input matrix/matrices

    Returns
    -------
    out : tree-like structure of jnp.ndarray (like M)
        Matrix logarithm of M
    """
    sig = "(n,n)->(n,n)"
    return tree_map(jnp.vectorize(_logm, signature=sig), M)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2013-2019 Max-Planck-Society
#
# NIFTy is being developed at the Max-Planck-Institut fuer Astrophysik.

from ..utilities import NiftyMeta


class Domain(metaclass=NiftyMeta):
    """The abstract class repesenting a (structured or unstructured) domain.
    """
    def __repr__(self):
        raise NotImplementedError

    def __hash__(self):
        """Returns a hash value for the object.

        Notes
        -----
        Only members that are explicitly added to
        :attr:`._needed_for_hash` will be used for hashing.
        """
        try:
            return self._hash
        except AttributeError:
            v = vars(self)
            self._hash = hash(tuple(v[key] for key in self._needed_for_hash))
        return self._hash

    def __eq__(self, x):
        """Checks whether two domains are equal.

        Parameters
        ----------
        x : Domain
            The domain `self` is compared to.

        Returns
        -------
        bool : True iff `self` and x describe the same domain.

        Notes
        -----
        Only members that are explicitly added to
        :attr:`._needed_for_hash` will be used for comparison.

        Subclasses of Domain should not re-define :meth:`__eq__`,
        :meth:`__ne__`, or :meth:`__hash__`; they should instead add their
        relevant attributes' names to :attr:`._needed_for_hash`.
        """
        if self is x:  # shortcut for simple case
            return True
        if not isinstance(x, type(self)):
            return False
        for key in self._needed_for_hash:
            if vars(self)[key] != vars(x)[key]:
                return False
        return True

    def __ne__(self, x):
        """Returns the opposite of :meth:`.__eq__()`"""
        return not self.__eq__(x)

    @property
    def shape(self):
        """tuple of int: number of pixels along each axis

        The shape of the array-like object required to store information
        defined on the domain.
        """
        raise NotImplementedError

    @property
    def size(self):
        """int: total number of pixels.

        Equivalent to the products over all entries in the domain's shape.
        """
        raise NotImplementedError



# Copyright(C) 2013-2021 Gordian Edenhofer
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from .lanczos import (
    lanczos_tridiag,
    stochastic_logdet_from_lanczos,
    stochastic_lq_logdet,
)
from .stats_distributions import (
    interpolator,
    invgamma_invprior,
    laplace_prior,
    invgamma_prior,
    lognormal_invprior,
    lognormal_moments,
    lognormal_prior,
    normal_invprior,
    normal_prior,
    uniform_prior,
)
from .unique import amend_unique, amend_unique_, unique



#!/usr/bin/env python3

# Copyright(C) 2013-2021 Max-Planck-Society
# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

# %% [markdown]
# # Demonstration of the non-parametric correlated field model in NIFTy.re

# ## The Model

# %%
import jax
import matplotlib.pyplot as plt
import nifty.re as jft
from jax import numpy as jnp
from jax import random

jax.config.update("jax_enable_x64", True)

seed = 42
key = random.PRNGKey(seed)

dims = (128, 128)

cf_zm = dict(offset_mean=0.0, offset_std=(1e-3, 1e-4))
cf_fl = dict(
    fluctuations=(1e-1, 5e-3),
    loglogavgslope=(-1.0, 1e-2),
    flexibility=(1e0, 5e-1),
    asperity=(5e-1, 5e-2),
)
cfm = jft.CorrelatedFieldMaker("cf")
cfm.set_amplitude_total_offset(**cf_zm)
cfm.add_fluctuations(
    dims, distances=1.0 / dims[0], **cf_fl, prefix="ax1", non_parametric_kind="power"
)
correlated_field = cfm.finalize()

scaling = jft.LogNormalPrior(3.0, 1.0, name="scaling", shape=(1,))


class Signal(jft.Model):
    def __init__(self, correlated_field, scaling):
        self.cf = correlated_field
        self.scaling = scaling
        # Init methods of the Correlated Field model and any prior model in
        # NIFTy.re are aware that their input is standard normal a priori.
        # The `domain` of a model does not know this. Thus, tracking the `init`
        # methods should be preferred over tracking the `domain`.
        super().__init__(init=self.cf.init | self.scaling.init)

    def __call__(self, x):
        # NOTE, think of `Model` as being just a plain function that takes some
        # input and performs all the necessary computation for your model.
        # Note, `scaling` here is completely degenarate with `offset_std` in the
        # likelihood but the priors for them are very different.
        return self.scaling(x) * jnp.exp(self.cf(x))


signal = Signal(correlated_field, scaling)

# %% [markdown]
# ## The likelihood

# %%
signal_response = signal
noise_cov = lambda x: 0.1**2 * x
noise_cov_inv = lambda x: 0.1**-2 * x

# Create synthetic data
key, subkey = random.split(key)
pos_truth = jft.random_like(subkey, signal_response.domain)
signal_response_truth = signal_response(pos_truth)
key, subkey = random.split(key)
noise_truth = (
    (noise_cov(jft.ones_like(signal_response.target))) ** 0.5
) * jft.random_like(key, signal_response.target)
data = signal_response_truth + noise_truth

lh = jft.Gaussian(data, noise_cov_inv).amend(signal_response)

# %% [markdown]
# ## The inference

# %%
n_vi_iterations = 6
delta = 1e-4
n_samples = 4

key, k_i, k_o = random.split(key, 3)
# NOTE, changing the number of samples always triggers a resampling even if
# `resamples=False`, as more samples have to be drawn that did not exist before.
samples, state = jft.optimize_kl(
    lh,
    jft.Vector(lh.init(k_i)),
    n_total_iterations=n_vi_iterations,
    n_samples=lambda i: n_samples // 2 if i < 2 else n_samples,
    # Source for the stochasticity for sampling
    key=k_o,
    # Names of parameters that should not be sampled but still optimized
    # can be specified as point_estimates (effectively we are doing MAP for
    # these degrees of freedom).
    # point_estimates=("cfax1flexibility", "cfax1asperity"),
    # Arguments for the conjugate gradient method used to drawing samples
    draw_linear_kwargs=dict(
        cg_name="SL",
        cg_kwargs=dict(absdelta=delta * jft.size(lh.domain) / 10.0, maxiter=100),
    ),
    # Arguements for the minimizer in the nonlinear updating of the samples
    nonlinearly_update_kwargs=dict(
        minimize_kwargs=dict(
            name="SN",
            xtol=delta,
            cg_kwargs=dict(name=None),
            maxiter=5,
        )
    ),
    # Arguments for the minimizer of the KL-divergence cost potential
    kl_kwargs=dict(
        minimize_kwargs=dict(
            name="M", xtol=delta, cg_kwargs=dict(name=None), maxiter=35
        )
    ),
    sample_mode="nonlinear_resample",
    odir="results_intro",
    resume=False,
)

# %%
namps = cfm.get_normalized_amplitudes()
post_sr_mean = jft.mean(tuple(signal(s) for s in samples))
post_a_mean = jft.mean(tuple(cfm.amplitude(s)[1:] for s in samples))
grid = correlated_field.target_grids[0]
to_plot = [
    ("Signal", signal(pos_truth), "im"),
    ("Noise", noise_truth, "im"),
    ("Data", data, "im"),
    ("Reconstruction", post_sr_mean, "im"),
    (
        "Amplitude spectrum",
        (
            grid.harmonic_grid.mode_lengths[1:],
            cfm.amplitude(pos_truth)[1:],
            post_a_mean,
        ),
        "loglog",
    ),
]
fig, axs = plt.subplots(2, 3, figsize=(16, 9))
for ax, v in zip(axs.flat, to_plot):
    title, field, tp, *labels = v
    ax.set_title(title)
    if tp == "im":
        end = tuple(n * d for n, d in zip(grid.shape, grid.distances))
        im = ax.imshow(field.T, cmap="inferno", extent=(0.0, end[0], 0.0, end[1]))
        plt.colorbar(im, ax=ax, orientation="horizontal")
    else:
        ax_plot = ax.loglog if tp == "loglog" else ax.plot
        x = field[0]
        for f in field[1:]:
            ax_plot(x, f, alpha=0.7)
for ax in axs.flat[len(to_plot) :]:
    ax.set_axis_off()
fig.tight_layout()
fig.savefig("results_intro_full_reconstruction.png", dpi=400)
plt.show()



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

import pprint
from typing import Any, NamedTuple, TypeVar

import jax
from jax import numpy as jnp
from jax.tree_util import tree_map

from .evi import Samples
from .tree_math import Vector, get_map

O = TypeVar("O")
I = TypeVar("I")


def _residual_params(inp):
    ndof = inp.size if jnp.isrealobj(inp) else 2 * inp.size
    mean = jnp.sum(inp) / inp.size
    rchisq = jnp.vdot(inp, inp).real / ndof
    return mean, rchisq, ndof


class ChiSqStats(NamedTuple):
    mean: Any
    reduced_chisq: Any
    ndof: Any


def reduced_residual_stats(position_or_samples, func=None, *, map="lmap"):
    """Computes the average, reduced chi-squared, and number of parameters
    as a summary statistics for a given input.

    Parameters:
    -----------
    position_or_samples: tree-like or Samples
        Input values to compute reduces chi-sq statistics. The statistics is
        computed for each leaf of the pytree, i.E. only array-like leafs are
        square averaged. If `positin_or_samples` is a `Sample` object, the
        chi-sq statistics is computed for each sample, and the sample mean and
        standard deviation of the statistics is returned.
    func: Callable (optional)
        Function to apply to `position_or_samples` before computing the chi-sq
        statistics for. If provided, the statistics is computed for `func(x)`
        instead of `x` where x is either primals or a sample.

    Returns:
    --------
    stats: tree-like
        Pytree of tuple containing the mean, reduced chi-squared, and number of
        parameters for each leaf of the input tree. For the mean and reduched
        chi-sq, a numpy array with the sample mean and sample std is returned.
        If samples is None, the second entry of this array is always zero.
    """
    map = get_map(map)
    if not isinstance(position_or_samples, Samples) or len(position_or_samples) == 0:
        if isinstance(position_or_samples, Samples):
            assert len(position_or_samples) == 0
            position_or_samples = position_or_samples.pos
        samples = tree_map(lambda x: x[jnp.newaxis, ...], position_or_samples)
    else:
        assert isinstance(position_or_samples, Samples)
        samples = position_or_samples.samples
    samples = map(func)(samples) if func is not None else samples

    get_stats = map(_residual_params)

    def red_chisq_stat(s):
        m, rx, nd = get_stats(s)
        m = jnp.array([jnp.mean(m), jnp.std(m)])
        rx = jnp.array([jnp.mean(rx), jnp.std(rx)])
        return ChiSqStats(m, rx, nd[0])

    return tree_map(red_chisq_stat, samples)


def _rpprint(ps: Any, *, _indent=0, _key="") -> str:
    if isinstance(ps, Vector):
        msg = _rpprint(ps.tree, _indent=_indent)
    elif isinstance(ps, dict):
        msg = ""
        for k, v in ps.items():
            k = _key + "/" * bool(_key) + str(k)
            if isinstance(v, dict):
                msg += _rpprint(v, _indent=_indent, _key=k)
            else:
                msg += "  " * _indent + f"{k:24s}::"
                m = _rpprint(v, _indent=_indent + 1, _key="")
                msg += (" " + m.lstrip()) if len(m.splitlines()) == 1 else ("\n" + m)
    elif isinstance(ps, (tuple, list)):
        msg = f"{'list' if isinstance(ps, list) else 'tuple'}(\n"
        for v in ps:
            msg += _rpprint(v, _indent=_indent + 1, _key="")
        msg += ")\n"
    else:
        # Catch all other instances with PrettyPrinter
        msg = "  " * _indent + pprint.pformat(ps) + "\n"
    return msg


def minisanity(position_or_samples, func=None, *, map="lmap"):
    """Wrapper for `reduced_residual_stats` to retrieve the reduced chi-squared
    and a pretty-printable string of the statistics."""
    stat_tree = reduced_residual_stats(position_or_samples, func=func, map=map)

    def make_pretty_string(x):
        rsq = x.reduced_chisq
        s = (
            f"reduced χ²:{rsq[0]:8.2}±{rsq[1]:8.2}"
            f", avg:{x.mean[0]:+9.2}±{x.mean[1]:8.2}"
            f", #dof:{int(x.ndof):7d}"
        )
        return s

    def is_leaf(l):
        return isinstance(l, ChiSqStats)

    ps = tree_map(make_pretty_string, stat_tree, is_leaf=is_leaf)
    return stat_tree, _rpprint(ps)



# SPDX-License-Identifier: GPL-2.0+ OR BSD-2-Clause

from jax import lax

_DISABLE_CONTROL_FLOW_PRIM = False


def cond(pred, true_fun, false_fun, operand):
    if _DISABLE_CONTROL_FLOW_PRIM:
        if pred:
            return true_fun(operand)
        else:
            return false_fun(operand)
    else:
        return lax.cond(pred, true_fun, false_fun, operand)


def while_loop(cond_fun, body_fun, init_val):
    if _DISABLE_CONTROL_FLOW_PRIM:
        val = init_val
        while cond_fun(val):
            val = body_fun(val)
        return val
    else:
        return lax.while_loop(cond_fun, body_fun, init_val)


def fori_loop(lower, upper, body_fun, init_val):
    if _DISABLE_CONTROL_FLOW_PRIM:
        val = init_val
        for i in range(int(lower), int(upper)):
            val = body_fun(i, val)
        return val
    else:
        return lax.fori_loop(lower, upper, body_fun, init_val)
